package de.credo.common.bundles.license.l4j;

import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.util.Date;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.eclipse.e4.core.contexts.ContextInjectionFactory;
import org.eclipse.e4.core.contexts.IEclipseContext;

import com.license4j.ActivationStatus;
import com.license4j.LicenseText;
import com.license4j.ValidationStatus;

import de.credo.common.bundles.license.License;
import de.credo.common.bundles.license.LicenseHandler;
import de.credo.common.db.customers.model.FeatureSets;
import de.credo.common.db.customers.model.Features;
import de.credo.common.db.customers.model.ProductEditions;
import de.credo.common.db.customers.model.ProductGroups;
import de.credo.common.db.customers.model.Products;
import jakarta.inject.Inject;

public class L4jLicense implements License {

	private	final static	Date	currentDate	= new Date();

	@Inject	LicenseHandler			licenseHandler;

	private	com.license4j.License	l4jLicense;

	private L4jLicense(com.license4j.License l4jLicense) {
		this.l4jLicense	= l4jLicense;
	}

	@Override
	public com.license4j.License getNative() {
		return l4jLicense;
	}

	@Override
	public LicenseTypes getLicenseType() {
		return isTrial() ? LicenseTypes.TRIAL : LicenseTypes.NODE_LOCKED;
	}

	@Override
	public long getLicenseId() {
		return l4jLicense.getLicenseText().getLicenseID();
	}

	@Override
	public String getLicenseKey() {
		return Long.toString(l4jLicense.getLicenseText().getLicenseID());
	}

	@Override
	public int getCustomerId() {
		String	customerId	= getLicenseFeatures().get("CID");
		return customerId != null ? Integer.parseInt(customerId) : 0;
	}

	public Map<String, String> getLicenseFeatures() {
		HashMap<String, String>	features	= new HashMap<>();
		features.putAll(l4jLicense.getLicenseText().getCustomSignedFeatures());
		features.putAll(l4jLicense.getLicenseText().getCustomUnsignedFeatures());

		return features;
	}

	@Override
	public String getProductId() {
		return l4jLicense.getLicenseText().getLicenseValidProductID();
	}

	public String getProductEdition() {
		return l4jLicense.getLicenseText().getLicenseValidProductEdition();
	}

	public String getProductVersion() {
		return l4jLicense.getLicenseText().getLicenseValidProductVersion();
	}

	@Override
	public Set<Features> getFeatures() {
		Set<Features>		features	= EnumSet.noneOf(Features.class);
		Map<String, String>	work		= new HashMap<>();

		// Erst Zusammenstellen einer einfachen Liste mit Features UND FeatureSets aus Defaults und Lizenzangaben
		if (getProduct().getDefaultFeatureSets() != null) {
			for (FeatureSets fs: getProduct().getDefaultFeatureSets()) {
				work.put(fs.name(), null);
			}
		}
		if (getProduct().getDefaultFeatures() != null) {
			for (Features f: getProduct().getDefaultFeatures()) {
				work.put(f.name(), null);
			}
		}
		work.putAll(getLicenseFeatures());

		// ... dann Auflösen dieser Liste hin zu einer reinen Feature-Liste

		// Default-Features der gefundenen FeatureSets
		for (FeatureSets fs: FeatureSets.fromStringMap(work)) {
			features.addAll(fs.getFeaturesResolved());
		}

		// Weitere in der Lizenz gefundene Features
		features.addAll(Features.fromStringMap(work));

		return features;
	}

	@Override
	public Products getProduct() {
		ProductGroups	productGroup	= ProductGroups.from(getProductId());
		ProductEditions	edition			= ProductEditions.from(getProductEdition());
		return Products.from(productGroup, edition);
	}

	@Override
	public boolean isEnabled() {
		return l4jLicense != null && l4jLicense.getActivationStatus() != ActivationStatus.ACTIVATION_REQUIRED;
	}

	@Override
	public boolean isValid() {
		return switch (l4jLicense.getValidationStatus()) {
			case LICENSE_VALID	->
				switch (l4jLicense.getActivationStatus()) {
					case ACTIVATION_COMPLETED,
							ACTIVATION_NOT_REQUIRED						-> true;
					case ACTIVATION_SERVER_CONNECTION_ERROR				-> false;
					case ACTIVATION_HARDWAREID_ERROR,
							ACTIVATION_NOT_FOUND_ON_SERVER,
							ACTIVATION_REJECTED_FEATURE_DISABLED,
							ACTIVATION_REJECTED_IP_BLOCK_RESTRICTION,
							ACTIVATION_REQUIRED,
							ALREADY_ACTIVATED_ON_ANOTHER_COMPUTER,
							DEACTIVATION_COMPLETED,
							DEACTIVATION_REJECTED_FEATURE_DISABLED,
							LICENSE_NOT_FOUND_ON_ACTIVATION_SERVER,
							MULTIPLE_ACTIVATION_LIMIT_REACHED,
							MULTIPLE_DEACTIVATION_LIMIT_REACHED			-> false;
					default												-> false;
				};
			case LICENSE_MAINTENANCE_EXPIRED,
					INCORRECT_SYSTEM_TIME,
					LICENSE_EXPIRED,
					LICENSE_INVALID,
					MISMATCH_HARDWARE_ID,
					MISMATCH_PRODUCT_EDITION,
					MISMATCH_PRODUCT_ID,
					MISMATCH_PRODUCT_VERSION,
					VALIDATION_REJECTED_FEATURE_DISABLED,
					VALIDATION_REJECTED_IP_BLOCK_RESTRICTION	-> false;
			default												-> false;
		};
	}

	@Override
	public boolean isTrial() {
		// Loopback an den LicenseHandler
		if (getLicenseId() == Long.parseLong(licenseHandler.getConfiguration().getProductIds().get(LicenseTypes.TRIAL).substring(
				licenseHandler.getConfiguration().getProductIds().get(LicenseTypes.TRIAL).indexOf("id: ") + 4,
				licenseHandler.getConfiguration().getProductIds().get(LicenseTypes.TRIAL).indexOf(")"))))
			return true;

		// Die Trial-Erkennung ist nur im Rahmen von < 100 Tagen Restlaufzeit interessant. Dieser Wert ist sowieso
		// nur bei normalen Lizenzen wichtig, da Floating Licenses immer ein Expire-Date zurückliefern.
		if (l4jLicense.getLicenseText().getLicenseFileType() == LicenseText.TYPE_LICENSE_FILE
				&& l4jLicense.getLicenseText().getLicenseExpireDate() != null)
			return true;

		return false;
	}

	@Override
	public boolean isExpired() {
		if (l4jLicense.getValidationStatus() == ValidationStatus.LICENSE_EXPIRED)
			return true;

		if (l4jLicense.getLicenseText().getLicenseExpireDate() != null
				&& l4jLicense.getLicenseText().getLicenseExpireDate().before(currentDate))
			return true;

		return false;
	}

	@Override
	public ZonedDateTime getLicenseExpireDate() {
		return l4jLicense.getLicenseText().getLicenseExpireDate() != null
				? ZonedDateTime.ofInstant(l4jLicense.getLicenseText().getLicenseExpireDate().toInstant(), ZoneOffset.UTC)
						: null;
	}

	@Override
	public int getLicenseExpireDaysRemaining(ZonedDateTime currentDate) {
		return l4jLicense.getLicenseText().getLicenseExpireDaysRemaining(currentDate == null ? new Date() : Date.from(currentDate.toInstant()));
	}

	@Override
	public boolean isMaintenanceExpired() {
		if (l4jLicense.getValidationStatus() == ValidationStatus.LICENSE_MAINTENANCE_EXPIRED)
			return true;

		if (l4jLicense.getLicenseText().getLicenseMaintenanceExpireDate() != null
				&& l4jLicense.getLicenseText().getLicenseMaintenanceExpireDate().before(currentDate))
			return true;

		return false;
	}

	@Override
	public ZonedDateTime getLicenseMaintenanceExpireDate() {
		return l4jLicense.getLicenseText().getLicenseMaintenanceExpireDate() != null
				? ZonedDateTime.ofInstant(l4jLicense.getLicenseText().getLicenseMaintenanceExpireDate().toInstant(), ZoneOffset.UTC)
				: null;
	}

	@Override
	public int getLicenseMaintenanceExpireDaysRemaining(ZonedDateTime currentDate) {
		return l4jLicense.getLicenseText().getLicenseMaintenanceExpireDaysRemaining(currentDate == null ? new Date() :Date.from(currentDate.toInstant()));
	}

	@Override
	public int getLicenseQuantity() {
		return l4jLicense.getLicenseText().getLicenseQuantity();
	}

	@Override
	public String getUserRegisteredTo() {
		return l4jLicense.getLicenseText().getUserRegisteredTo();
	}

	@Override
	public String getUserCity() {
		return l4jLicense.getLicenseText().getUserCity();
	}

	@Override
	public String getUserCountry() {
		return l4jLicense.getLicenseText().getUserCountry();
	}

	public static License make(com.license4j.License l4jLicense, IEclipseContext eclipseContext) {
		License	newLicense	= new L4jLicense(l4jLicense);

		if (eclipseContext != null) {
			ContextInjectionFactory.inject(newLicense, eclipseContext);
		}

		return newLicense;
	}

	public ValidationStatus getValidationStatus() {
		return l4jLicense.getValidationStatus();
	}

	public ActivationStatus getActivationStatus() {
		return l4jLicense.getActivationStatus();
	}

}
