package de.credo.common.bundles.license.l4j;

import java.io.BufferedReader;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.ICoreRunnable;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.e4.core.contexts.ContextInjectionFactory;
import org.eclipse.e4.core.contexts.IEclipseContext;
import org.eclipse.e4.core.di.annotations.Optional;
import org.eclipse.e4.core.services.nls.Translation;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Shell;
import org.osgi.service.component.annotations.Component;
import org.osgi.service.prefs.BackingStoreException;

import com.license4j.ActivationStatus;
import com.license4j.LicenseValidator;
import com.license4j.ValidationStatus;

import de.credo.common.bundles.cmd.CmdArgument;
import de.credo.common.bundles.license.License;
import de.credo.common.bundles.license.License.LicenseTypes;
import de.credo.common.bundles.license.LicenseHandler;
import de.credo.common.bundles.license.l4j.prefpages.EnterLicenseDialog;
import de.credo.common.bundles.license.l4j.prefpages.LicenseUsageDialog;
import de.credo.common.bundles.proxy.ProxyData;
import de.credo.common.bundles.version.BundleVersion;
import jakarta.inject.Inject;
import jakarta.inject.Singleton;

@Component @Singleton
public class L4jLicenseHandler implements LicenseHandler {

	private	static final	String	LICENSE_FILE_ARG				= "-licenseFile";

	private	static final	String	TECH_LICENSE_SERVER_ADDRESS		= "https://license.cre-do.de/algas";

	private	final static	String	CONFIG_KEY_LICENSE_SERVER		= "licenseServer";
	private	final static	String	CONFIG_KEY_ORIGINAL_LICENSE		= "originalLicense";
	private	final static	String	CONFIG_KEY_ACTIVATED_LICENSE	= "activatedLicense";
	private	final static	String	CONFIG_KEY_LAST_ACTIVATION		= "lastActivation";
	private	final static	String	CONFIG_KEY_SECURE_HASH			= "security";

	@Inject	@Translation	Messages	messages;

	private	String					licenseServer					= TECH_LICENSE_SERVER_ADDRESS;
	private	com.license4j.License	originalLicense;
	private	com.license4j.License	activatedLicense;
	private	long					lastActivationEpoch;

	private	LicenseHandlerConfiguration	configuration;

	public class L4jConfiguration implements LicenseHandlerConfiguration {

		private	IEclipseContext				context;
		private	Map<LicenseTypes, String>	productIds;
		private	String						productName;
		private	String						publicKey;

		@Override
		public IEclipseContext getContext() {
			return context;
		}

		@Override
		public LicenseHandlerConfiguration setContext(IEclipseContext context) {
			this.context = context;
			return this;
		}

		@Override
		public Map<LicenseTypes, String> getProductIds() {
			return productIds;
		}

		@Override
		public LicenseHandlerConfiguration setProductIds(Map<LicenseTypes, String> productIds) {
			this.productIds	= productIds;
			return this;
		}

		@Override
		public String getProductName() {
			return productName;
		}

		@Override
		public LicenseHandlerConfiguration setProductName(String productName) {
			return this;
		}

		@Override
		public String getPublicKey() {
			return publicKey;
		}

		@Override
		public LicenseHandlerConfiguration setPublicKey(String publicKey) {
			this.publicKey	= publicKey;
			return this;
		}
	}

	@Override
	public LicenseHandlerConfiguration configBuilder() {
		return new L4jConfiguration();
	}

	@Override
	public void init(LicenseHandlerConfiguration configuration) {
		Assert.isNotNull(configuration);
		Assert.isNotNull(configuration.getContext());
		Assert.isNotNull(configuration.getProductIds());
		Assert.isTrue(!configuration.getProductIds().isEmpty());
		this.configuration	= configuration;

		ContextInjectionFactory.inject(this, configuration.getContext());
	}

	@Override
	public LicenseHandlerConfiguration getConfiguration() {
		return configuration;
	}

	@Override
	public boolean readLicense() {
		boolean	success;
		if (success = loadPreferences())
			return success;

		// Lizenzlesen nicht erfolgreich? Dann gibt es vielleicht einen passenden Lizenzparameter -licenseFile ...?
		if (!success) {
			String[]	args	= Platform.getApplicationArgs();
			int			idx		= 0;
			for (idx = 0; idx < args.length; idx++) {
				if (args[idx].equalsIgnoreCase(LICENSE_FILE_ARG)) {
					break;
				}
			}

			// Parameter gefunden? Dann Auswerten und Versuch einer Validierung (-1 weil ja noch der eigentliche Wert kommen muss
			if (idx < args.length - 1) {
				if (args[idx].equalsIgnoreCase(LICENSE_FILE_ARG)) {
					// Routine für übergebene Lizenzdatei
					Path	licenseFile	= Paths.get(args[idx + 1]);
					if (Files.exists(licenseFile)) {
						try(BufferedReader	br	= Files.newBufferedReader(licenseFile)) {
							// Einlesen der Lizenzdatei
							String	content	= br.lines().collect(Collectors.joining("\n"));
							// Check, ob diese gültig ist
							setNewLicenseString(content);
							if (success = originalLicense != null) {
								if (!(success = !originalLicense.isActivationRequired())) {
									// Aktivieren, wenn nötig
									activateLicense(null);
									success	= activatedLicense.isActivationCompleted();
								}
							}
						} catch (IOException e) {
							MessageDialog.openError(null, messages.licenseHandlerDialogTitleError, messages.licenseHandlerDialogProblemWhileReadingFile
									.replace("%1", licenseFile.toAbsolutePath().toString())
									.replace("%2", e.getLocalizedMessage()));
						}
					}
				}
			}
		}

		return success;
	}

	@Override
	public boolean checkLicense() {
		boolean	checkResult		= false;
		com.license4j.License	checkedLicense	= null;

		if (originalLicense != null) {
			ValidationStatus	originalState	= checkValidationStatus(originalLicense);
			if (originalState == ValidationStatus.LICENSE_VALID || originalState == ValidationStatus.LICENSE_MAINTENANCE_EXPIRED) {
				if (activatedLicense != null) {
					if (validateActivatedLicense()) {
						checkResult		= true;
						checkedLicense	= activatedLicense;

					} else if (activatedLicense.getActivationStatus() == com.license4j.ActivationStatus.ACTIVATION_SERVER_CONNECTION_ERROR) {
						checkResult		= false;
						checkedLicense	= originalLicense;

					} else {
						// OriginalLicense, weil die ActivatedLicense ja sowieso bereits aktiviert ist und IMMER false zurückliefert!
						checkResult		= !originalLicense.isActivationRequired();
						checkedLicense	= activatedLicense;
					}

				} else {
					checkResult		= !originalLicense.isActivationRequired();
					checkedLicense	= originalLicense;
				}

			} else {
				// Tue nichts. Die gespeicherte Original-Lizenz darf auch nicht gelöscht werden, weil sich Stati
				// wie INCORRECT_SYSTEM_TIME (nur als Beispiel) bis zum nächsten Mal ja noch einmal ändern könnten.
			}

		// Na gut, dann gibt es halt noch keine gespeicherte Lizenz. Das ist auch eine wertvolle Information (siehe LifeCycle-Class).
		// Dann ist trotzdem eine null-Instanz zu registrieren, weil sonst das ein oder andere Inject auf die Nase fällt!
		} else {
			checkResult		= false;
			checkedLicense	= null;
		}

		if (checkResult) {
			persistPreferences();
//			showLicenseMessageIfAny(checkedLicense);
		}

		// Hat es ein Problem mit der Lizenz gegeben, dann wird einfach keine in den Kontext gelegt, viel mehr eine dort ggfs. noch vorhandene gelöscht.
		configuration.getContext().set(License.class, checkedLicense == null ? null : L4jLicense.make(checkedLicense, configuration.getContext()));

		return checkResult;
	}

	protected com.license4j.License toLicense(String licenseString) {
		if (licenseString == null)
			return null;

		// Ermittlung des letzten Build-Datums, um Prüfungen auf abgelaufene Maintenance zu ermöglichen
		Date	releaseDate	= null;
		if (BundleVersion.getBuildTime() != null) {
			releaseDate	= Date.from(BundleVersion.getBuildTime());
		}

		return LicenseValidator.validate(licenseString, configuration.getPublicKey(), configuration.getProductIds().get(LicenseTypes.NODE_LOCKED), null, null, new Date(), releaseDate);
	}

	@Override
	public void openEnterLicenseDialog(Shell parentShell) {
		EnterLicenseDialog	dialog	= new EnterLicenseDialog(parentShell);
		ContextInjectionFactory.inject(dialog, configuration.getContext());
		dialog.setBlockOnOpen(true);

		if (dialog.open() == Dialog.OK) {
			setNewLicenseString(dialog.getEnteredLicenseKey());
		}
	}

	@Override
	public void openLicenseUsageDialog(Shell parentShell) {
		LicenseUsageDialog	dialog	= new LicenseUsageDialog(parentShell);
		ContextInjectionFactory.inject(dialog, configuration.getContext());
		dialog.open();
	}

	ValidationStatus checkValidationStatus(com.license4j.License license) {
		com.license4j.ValidationStatus	status	= license.getValidationStatus();
		switch (status) {
		case LICENSE_VALID:
			break;

		case INCORRECT_SYSTEM_TIME:
			MessageDialog.openError(null, messages.licenseHandlerDialogTitleError, messages.licenseHandlerDialogIncorrectSystemTime);
			break;

		case LICENSE_EXPIRED:
			MessageDialog.openError(null, messages.licenseHandlerDialogTitleError, messages.licenseHandlerDialogLicenseExpired);
			break;

		case LICENSE_INVALID:
			MessageDialog.openError(null, messages.licenseHandlerDialogTitleError, messages.licenseHandlerDialogLicenseInvalid);
			break;

		case MISMATCH_HARDWARE_ID:
			MessageDialog.openError(null, messages.licenseHandlerDialogTitleError, messages.licenseHandlerDialogMismatchHardwareId);
			break;

		case MISMATCH_PRODUCT_EDITION:
		case MISMATCH_PRODUCT_ID:
		case MISMATCH_PRODUCT_VERSION:
			MessageDialog.openError(null, messages.licenseHandlerDialogTitleError, messages.licenseHandlerDialogMismatchProduct.replace("%1", configuration.getProductIds().get(LicenseTypes.NODE_LOCKED)));
			break;

		case LICENSE_MAINTENANCE_EXPIRED:
			break;
		case VALIDATION_REJECTED_FEATURE_DISABLED:
			break;
		case VALIDATION_REJECTED_IP_BLOCK_RESTRICTION:
			break;
		default:
			MessageDialog.openError(null, messages.licenseHandlerDialogTitleError, messages.licenseHandlerDialogUnkownValidationStatus.replace("%1", status.toString()));
			break;
		}

		return ValidationStatus.valueOf(status.name());
	}

	private boolean validateActivatedLicense() {
		loadPreferences();
		Instant	lastActivationDate	= Instant.ofEpochSecond(lastActivationEpoch);

		// Jede Woche wird versucht, die Lizenz erneut zu aktivieren, als erweiterter Lizenzcheck
		// Zu diesem Zeitpunkt noch heimlich
		if (lastActivationDate.plus(7, ChronoUnit.DAYS).isBefore(Instant.now()))
			return switch (LicenseValidator.checkOnlineAvailability(configuration.getPublicKey(), activatedLicense, getTechLicenseServer(), 2000)) {
			case 1:		// Alles ok
 {
				updateLastActivationDate();
				yield true;
			}
			case -1:	// Keine Verbindung zum Server möglich, jetzt kein Problem
				yield true;
			default:
				yield false;
			};
		else if (lastActivationDate.plus(90, ChronoUnit.DAYS).isBefore(Instant.now()))
			return switch (LicenseValidator.checkOnlineAvailability(configuration.getPublicKey(), activatedLicense, getTechLicenseServer(), 2000)) {
			case 1:		// Alles ok
 {
				updateLastActivationDate();
				yield true;
			}
			case -1:	// Keine Verbindung zum Server möglich, hier ist das ein Problem
				yield false;
			default:
				yield false;
			};
		else if (lastActivationDate.isBefore(Instant.now()))
			return activatedLicense.getActivationStatus() == com.license4j.ActivationStatus.ACTIVATION_COMPLETED
				|| activatedLicense.getActivationStatus() == com.license4j.ActivationStatus.ACTIVATION_NOT_REQUIRED;

		else
			return false;
	}

	public ActivationStatus activateOriginalLicense() {
		activatedLicense	= LicenseValidator.autoActivate(originalLicense, getTechLicenseServer());
		com.license4j.ActivationStatus	status	= activatedLicense.getActivationStatus();

		switch (status) {
		case ACTIVATION_REQUIRED:
			// Na gut! Dann aktiviere ich die Lizenz halt NOCH MAL ...
			activatedLicense		= LicenseValidator.autoActivate(activatedLicense, getTechLicenseServer());
			status					= activatedLicense.getActivationStatus();
			// Absichtlicher Überlauf in den SUCCESS-Fall

		case ACTIVATION_NOT_REQUIRED:
		case ACTIVATION_COMPLETED:
			if (activatedLicense.isActivationCompleted()) {
				updateLastActivationDate();
			}
			return ActivationStatus.ACTIVATION_COMPLETED;

		case ACTIVATION_HARDWAREID_ERROR:
			MessageDialog.openError(null, messages.licenseHandlerDialogTitleError, messages.licenseHandlerDialogActivationHardwareIdError);
			break;

		case LICENSE_NOT_FOUND_ON_ACTIVATION_SERVER:
		case ACTIVATION_NOT_FOUND_ON_SERVER:
			MessageDialog.openError(null, messages.licenseHandlerDialogTitleError, messages.licenseHandlerDialogActivationNotFoundOnServer);
			break;

		case ALREADY_ACTIVATED_ON_ANOTHER_COMPUTER:
			MessageDialog.openError(null, messages.licenseHandlerDialogTitleError, messages.licenseHandlerDialogAlreadyActivatedOnAnother);
			break;

		case MULTIPLE_ACTIVATION_LIMIT_REACHED:
			MessageDialog.openError(null, messages.licenseHandlerDialogTitleError, messages.licenseHandlerDialogMultipleActivationLimit);
			break;

		case ACTIVATION_SERVER_CONNECTION_ERROR:
			MessageDialog.openError(null, messages.licenseHandlerDialogTitleError, messages.licenseHandlerDialogActivationServerConnectionError);
			break;

		case ACTIVATION_REJECTED_FEATURE_DISABLED:
		case ACTIVATION_REJECTED_IP_BLOCK_RESTRICTION:
		case DEACTIVATION_COMPLETED:
		case DEACTIVATION_REJECTED_FEATURE_DISABLED:
		case MULTIPLE_DEACTIVATION_LIMIT_REACHED:
		default:
			MessageDialog.openError(null, messages.licenseHandlerDialogTitleError, messages.licenseHandlerDialogUnkownActivationStatus.replace("%1", status.toString()));
			break;
		}

		activatedLicense	= null;
		return ActivationStatus.valueOf(status.name());
	}

	public ActivationStatus deactivateActivatedLicense() {
		activatedLicense	= LicenseValidator.autoDeactivate(activatedLicense, getTechLicenseServer());
		com.license4j.ActivationStatus	status	= activatedLicense.getActivationStatus();

		switch (status) {
		case LICENSE_NOT_FOUND_ON_ACTIVATION_SERVER:
		case ACTIVATION_NOT_FOUND_ON_SERVER:
		case DEACTIVATION_COMPLETED:
			if (activatedLicense.isActivationCompleted()) {
				updateLastActivationDate();
			}
			activatedLicense	= null;
			return ActivationStatus.DEACTIVATION_COMPLETED;

		case ACTIVATION_SERVER_CONNECTION_ERROR:
			MessageDialog.openError(null, messages.licenseHandlerDialogTitleError, messages.licenseHandlerDialogDeactivationServerConnectionError);
			break;

		case ACTIVATION_HARDWAREID_ERROR:
		case ALREADY_ACTIVATED_ON_ANOTHER_COMPUTER:
		case MULTIPLE_ACTIVATION_LIMIT_REACHED:
		case ACTIVATION_NOT_REQUIRED:
		case ACTIVATION_REJECTED_FEATURE_DISABLED:
		case ACTIVATION_REJECTED_IP_BLOCK_RESTRICTION:
		case ACTIVATION_COMPLETED:
		case ACTIVATION_REQUIRED:
		case DEACTIVATION_REJECTED_FEATURE_DISABLED:
		case MULTIPLE_DEACTIVATION_LIMIT_REACHED:
		default:
			MessageDialog.openError(null, messages.licenseHandlerDialogTitleError, messages.licenseHandlerDialogUnkownDeactivationStatus.replace("%1", status.toString()));
		}

		activatedLicense	= null;
		return ActivationStatus.valueOf(status.name());
	}

	private void updateLastActivationDate() {
		lastActivationEpoch	= Instant.now().getEpochSecond();
		persistPreferences();

		// Update der Nutzungsstatistik auf dem Server nachgelagert, muss die weitere Verarbeitung nicht ausbremsen
		Job.create("Update license usage", (ICoreRunnable) arg0 ->
			LicenseValidator.updateLicenseUseInfo(configuration.getPublicKey(), originalLicense, getTechLicenseServer(), 2000))
		.schedule();
	}

	public void setNewLicenseString(String licenseString) {
		// Extrahieren der eigentlichen Lizenz, falls der Kunde mehr eingegeben hat, als er sollte
		// Beginn ermitteln...
		int	licenseStart	= licenseString.indexOf("# ");
		if (licenseStart > -1) {
			licenseString	= licenseString.substring(licenseStart);
			int		cx		= -1;
			char[]	olChars	= licenseString.toCharArray();
			char[]	nlChars	= new char[licenseString.length()];

			for (int ix = 0; ix < olChars.length; ix++) {
				if (olChars[ix] == '\r') {			// CR, Wird später sowieso unterdrückt
					// Tue nichts

				} else if (olChars[ix] == '\n') {	// LF, Zeilenumbruch
					// Entfernen von Sonderzeichen am Zeilenende sowie Leerzeilen
					while (nlChars[cx] == ' ' || nlChars[cx] == '\n') {
						cx--;
					}
					nlChars[++cx]	= olChars[ix];

				} else if (olChars[ix] == '<') {	// <, Beginn des Textes "</Key>", damit implizites Ende der Verarbeitung
					break;

				} else {
					nlChars[++cx]	= olChars[ix];
				}
			}

			// Prüfen, ob die eingegebene Lizenz grundsätzlich gültig ist und ggfs. gleich aktivieren
			originalLicense		= toLicense(new String(Arrays.copyOf(nlChars, cx + 1)).trim());
			activatedLicense	= null;
			checkLicense();

		} else {
			MessageDialog.openError(null, messages.licenseHandlerDialogTitleError, messages.licenseHandlerDialogNoValidLicenseKey);
		}
	}

	@Override
	public void activateLicense(Object parm) {
		activateOriginalLicense();
		persistPreferences();
		configuration.getContext().set(License.class, activatedLicense == null ? null : L4jLicense.make(activatedLicense, configuration.getContext()));
	}

	@Override
	public void deactivateLicense(Object parm) {
		deactivateActivatedLicense();
		persistPreferences();
		configuration.getContext().set(License.class, null);		// activatedLicense == null
	}

	@Override
	public String checkForUpdateVersion() {
		return originalLicense.checkForUpdate(null, BundleVersion.getVersionNumber(), getTechLicenseServer());
	}

	@Override
	public String checkForUpdateMessage(String updateVersion) {
		return originalLicense.checkForUpdateMessage(null, updateVersion, getTechLicenseServer());
	}

	@Override
	public String[] checkForNewMessage() {
		return originalLicense.checkForNewMessage(getTechLicenseServer());
	}

	@Override
	public String getTechLicenseServer() {
		return licenseServer;
	}

	@Override
	public void setTechLicenseServer(String licenseServer) {
		this.licenseServer = licenseServer;
	}

	public com.license4j.License getOriginalLicense() {
		return originalLicense;
	}

	public void setOriginalLicense(com.license4j.License originalLicense) {
		this.originalLicense = originalLicense;
	}

	public com.license4j.License getActivatedLicense() {
		return activatedLicense;
	}

	public void setActivatedLicense(com.license4j.License activatedLicense) {
		this.activatedLicense = activatedLicense;
	}

	public long getLastActivationEpoch() {
		return lastActivationEpoch;
	}

	public void setLastActivationEpoch(long lastActivationEpoch) {
		this.lastActivationEpoch = lastActivationEpoch;
	}

	@Override
	public CmdArgument[] commandLineArguments() {
		return new CmdArgument[] {
				new CmdArgument(LICENSE_FILE_ARG, true, true, messages.licenseHandlerCmdArgLicenseFile)};
	}

	public ValidationStatus checkValidationStatus(License license) {
		return null;
	}

	@Inject
	public void setProxyData(@Optional ProxyData proxyData) {
		if (proxyData != null) {
			// Wurde ein spezieller User-Agent gesetzt? Dann muss dieser auch in die L4J-spezifische Variable übernommen werden,
			// damit sowohl die eigenen Routinen als auch die ALGAS-Abfrage mit dem richtig falschen User-Agent läuft.
			if (proxyData.getUserAgent() != null) {
// Funktioniert nicht. Egal, welcher String hier außer dem impliziten Standardwert "License4J HTTPClient" gesetzt wird,
// wird am Ende vom ALGAS abgelehnt mit mit einem ActivationStatus.ACTIVATION_SERVER_CONNECTION_ERROR beantwortet. :-(
// Jedenfalls vermute ich den Fehler dort.
//				System.setProperty("license4j-user-agent", proxyData.userAgent);
			}
		}
	}

	public boolean loadPreferences() {
		setTechLicenseServer(globalPreferences.get(CONFIG_KEY_LICENSE_SERVER, getTechLicenseServer()));

		setOriginalLicense(toLicense(globalPreferences.get(CONFIG_KEY_ORIGINAL_LICENSE, null)));
		setActivatedLicense(toLicense(globalPreferences.get(CONFIG_KEY_ACTIVATED_LICENSE, null)));

		if (getOriginalLicense() != null) {
			if (getActivatedLicense() == null)
				return false;

			// Normales Ende einer bereits lizensierten Installation
			setLastActivationEpoch(globalPreferences.getLong(CONFIG_KEY_LAST_ACTIVATION, 0));
			String	secureHashString	= globalPreferences.get(CONFIG_KEY_SECURE_HASH, "");
			return secureHashString.equals(generateSecureHashString());

		} else if (Arrays.asList(Platform.getApplicationArgs()).contains(LICENSE_FILE_ARG))
			return false;		// Detaillierte Auswertung dieses Parameters in der globalen Routine

		else {
			// Testlizenz vorbelegen
			setOriginalLicense(toLicense(configuration.getProductIds().get(LicenseTypes.TRIAL)));
			configuration.getContext().set(License.class, L4jLicense.make(getOriginalLicense(), configuration.getContext()));
			return true;
		}
	}

	public void persistPreferences() {
		try {
			globalPreferences.put(CONFIG_KEY_LICENSE_SERVER, getTechLicenseServer());

			if (getOriginalLicense() != null && getOriginalLicense().getLicenseString() != null) {
				globalPreferences.put(CONFIG_KEY_ORIGINAL_LICENSE, getOriginalLicense().getLicenseString());
			} else {
				globalPreferences.remove(CONFIG_KEY_ORIGINAL_LICENSE);
			}

			if (getActivatedLicense() != null && getActivatedLicense().getLicenseString() != null) {
				globalPreferences.put(CONFIG_KEY_ACTIVATED_LICENSE, getActivatedLicense().getLicenseString());
			} else {
				globalPreferences.remove(CONFIG_KEY_ACTIVATED_LICENSE);
			}

			globalPreferences.putLong(CONFIG_KEY_LAST_ACTIVATION, getLastActivationEpoch());

			globalPreferences.put(CONFIG_KEY_SECURE_HASH, generateSecureHashString());
			globalPreferences.flush();
		} catch (BackingStoreException e) {
		}
	}

	public String generateSecureHashString() {
		List<String>	objectsToHash	= new ArrayList<>();
		String			hash			= "";

		if (getOriginalLicense() != null) {
			objectsToHash.add(getOriginalLicense().getLicenseString());
		}
		if (getActivatedLicense() != null) {
			objectsToHash.add(getActivatedLicense().getLicenseString());
		}
		objectsToHash.add(Long.toString(getLastActivationEpoch()));

		for (String hs: objectsToHash) {
			if (hs != null) {
				hash	+= Integer.toHexString(hs.hashCode());
			}
		}

		hash = "1" + hash;
		if (hash.length() % 2 > 0) {
			hash += "f";
		}
		return hash;
	}

}
