package de.credo.lookup2;

import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.core.runtime.jobs.ProgressProvider;
import org.eclipse.e4.core.contexts.ContextInjectionFactory;
import org.eclipse.e4.core.contexts.IEclipseContext;
import org.eclipse.e4.core.services.events.IEventBroker;
import org.eclipse.e4.ui.di.UISynchronize;
import org.eclipse.e4.ui.model.application.MApplication;
import org.eclipse.e4.ui.model.application.ui.basic.MTrimmedWindow;
import org.eclipse.e4.ui.workbench.modeling.EModelService;
import org.eclipse.jface.resource.FontDescriptor;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.RGB;
import org.osgi.service.event.Event;
import org.osgi.service.event.EventHandler;

import de.credo.lookup2.plugins.api.L2Constants;
import de.credo.lookup2.plugins.api.L2FileData;
import de.credo.lookup2.plugins.api.L2ProgressMonitor;
import de.credo.lookup2.plugins.impl.FileData;

class AppStartupCompleteObserver implements EventHandler {

	private	final static	int	CHECK_DELAY	= 2500;		// in Millisekunden

	private	final	EModelService		modelService;
	private	final	MApplication		application;
	private	final	IEclipseContext		appContext;
	private	final	IEventBroker		eventBroker;
	private	final	GlobalAppManager	globalAppManager;

	private	MTrimmedWindow		mainWindow;

	public AppStartupCompleteObserver(EModelService modelService, MApplication application, IEventBroker eventBroker) {
		this.modelService	= modelService;
		this.application	= application;
		this.eventBroker	= eventBroker;
		appContext			= application.getContext();
		globalAppManager	= ContextInjectionFactory.make(GlobalAppManager.class, appContext);
		appContext.set(GlobalAppManager.class, globalAppManager);
	}

	@Override
	public void handleEvent(Event event) {
		// Ein paar Inits
		JFaceResources.getColorRegistry().put(L2Constants.CREDO_PETROL_COLOR_KEY, new RGB(0, 126, 158));		// #007e9e
		JFaceResources.getColorRegistry().put(L2Constants.CREDO_GRAY_COLOR_KEY, new RGB(69, 76, 84));			// #454c54
		JFaceResources.getFontRegistry().put(L2Constants.BOLD_ITALIC_FONT_KEY, FontDescriptor.createFrom(JFaceResources.getDefaultFont()).setStyle(SWT.ITALIC | SWT.BOLD).getFontData());

		Job.getJobManager().setProgressProvider(new ProgressProvider() {

			@Override
			public IProgressMonitor createMonitor(Job job, IProgressMonitor group, int ticks) {
				try {
					return application.getSelectedElement().getContext().get(L2ProgressMonitor.class).addJob(job);
				} catch (NullPointerException e) {
					return null;
				}
			}

			@Override
			public IProgressMonitor createMonitor(Job job) {
				try {
					return application.getSelectedElement().getContext().get(L2ProgressMonitor.class).addJob(job);
				} catch (NullPointerException e) {
					return null;
				}
			}
		});

		// Liste von Dateien aus der Kommandozeile lesen
		List<Path>	files	= getFilesFromCommandLine();

		// Erstes Fenster aus Snippets erzeugen. Das ist zukünftig das Hauptfenster.
		mainWindow	= (MTrimmedWindow) modelService.cloneSnippet(application, Constants.MAIN_WINDOW, null);

		// Fensterposition und -ausdehnung wiederherstellen
		LifeCycle.restoreMainWindowPosition(modelService, mainWindow);

		// Erst jetzt wird das eigentliche Hauptfenster erzeugt! Dieser Befehl löst dummerweise auch den
		// PerspectiveChangedListener aus, was toll wäre, würden nicht die Perspective Toolbars erst in den
		// folgenden Schritten initialisiert. Der Schritt, die Start-Perspective-Toolbar wieder her-
		// zustellen, muss daher ganz unten noch einmal wiederholt werden.
		application.getChildren().add(mainWindow);

		// Startperspektive setzen
		globalAppManager.switchPerspective(mainWindow, globalAppManager.getStartPerspective());

		// Ausblenden aller Viewer-Toolbars und -Menüs, da zum Start grundsätzlich kein Viewer aktiv ist
		ToolbarUtils.hideViewerToolbars(mainWindow);
// TODO	ToolbarUtils.hideViewerMenus(mainWindow);
		ToolbarUtils.hideAllMenus(mainWindow);

		// Ausblenden aller Parser-Toolbars und -Menüs, da zum Start grundsätzlich kein Parser aktiv ist
// TODO	ToolbarUtils.hideParserToolbars(mainWindow);
// TODO	ToolbarUtils.hideParserMenus(mainWindow);

		// Ausblenden aller Perspektiven-Toolbars und -Menüs und zurückholen der dummerweise mit ausgeblendeten Perspective-Toolbar der Startperspektive (sofern vorhanden)
		ToolbarUtils.hidePerspectiveToolbars(mainWindow);
		ToolbarUtils.showPerspectiveToolbars(mainWindow, globalAppManager.determinePerspectiveId(mainWindow, globalAppManager.getStartPerspective()), true);
// TODO	ToolbarUtils.hidePerspectiveMenus(mainWindow);

		// Anlegen eines periodischen Jobs, in dem die weitere regelmäßige Kontrolle eines
		// Übergabefiles zwischen zwei Instanzen abgehandelt wird

		// Schleife zum regelmäßigen Prüfen auf weitere zu öffnende Dateien
		Timer	checkForStartupFileTimer	= new Timer("Refresh Open Files");
		checkForStartupFileTimer.schedule(new TimerTask() {

			@Override
			public void run() {
				if (Files.exists(StartupFileHandler.getStartupFile())) {
					UISynchronize	syncronizer	= appContext.get(UISynchronize.class);
					syncronizer.asyncExec(() -> {
						// Lesen der Startup-Datei und Öffnen der Dateien
						openFiles(StartupFileHandler.readAndDelete());
					});
				}
			}
		}, CHECK_DELAY, CHECK_DELAY);

		// Lesen der Kommandozeile und ggfs. Öffnen der übergebenen Dateien
		openFiles(files);
	}

	private List<Path> getFilesFromCommandLine() {
		String[]	arguments	= Platform.getApplicationArgs();
		boolean		ignoreNext	= false;
		List<Path>	files		= new ArrayList<>();

		for (String arg: arguments) {
			// Grundsätzliche alle Parameter beginnend mit '-' und die folgenden Werte sind zu unterdrücken.
			// Achtung: Es gibt auch Parameter ohne zusätzliche Werte!
			if (arg.charAt(0) == '-' || arg.charAt(0) == '/') {
				ignoreNext	= true;
			} else {
				if (ignoreNext) {
					ignoreNext	= false;
				} else {
					java.nio.file.Path	p	= Paths.get(arg);
					if (Files.exists(p)) {
						files.add(p);
					}
				}
			}
		}

		return files;
	}

	private void openFiles(List<Path> files) {
		// Diese Routine setzt jetzt endlich auch die richtige Perspektive. Wurde keine Datei übergeben,
		// dann passiert das im else-Fall für die Start-Perspektive. Im Fall einer sofort zu öffnenden
		// Datei wird per .openInWindow() einfach gleich die richtige Perspektive und -Toolbar gesetzt,
		// um unnötiges Flackern zu vermeiden.
		if (!files.isEmpty()) {
			// Sehr ähnliche Routine in OpenFileHandler!!
			for (int i = 0; i < files.size(); i++) {
				if (i == 0
						&& application.getChildren().size() > 1
						&& application.getChildren().get(1).getContext().get(L2FileData.class) == null) {		// Index 1, weil das Fenster mit Index 0 das versteckte erste Programmfenster ist
					// Erste Datei im aktuellen Fenster ...
					globalAppManager.openInWindow(mainWindow, new FileData(files.get(i), Files.isWritable(files.get(i))), SwitchPerspectiveTargetTypes.ANY);
				} else {
					// ... und weitere Dateien in weiteren, zusätzlichen Fenstern öffnen
					globalAppManager.openInNewWindow(new FileData(files.get(i), Files.isWritable(files.get(i))), true, SwitchPerspectiveTargetTypes.ANY);
				}
			}
		} else {
			// Die Start-Perspektive ist gesetzt, jetzt muss auch noch - oder nur noch - der Button gesetzt werden
			Map<Class<?>, Object>	values	= new HashMap<>(3);
			values.put(MTrimmedWindow.class, mainWindow);
			values.put(String.class, globalAppManager.determinePerspectiveId(mainWindow, globalAppManager.getStartPerspective()));
			eventBroker.send(Constants.SET_PERSPECTIVE_BUTTON_EVENT, values);
		}
	}

}
