package de.credo.lookup2;

import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.stream.Collectors;

import org.eclipse.core.runtime.preferences.IEclipsePreferences;
import org.eclipse.core.runtime.preferences.InstanceScope;
import org.eclipse.e4.core.contexts.Active;
import org.eclipse.e4.core.contexts.ContextInjectionFactory;
import org.eclipse.e4.core.di.annotations.Creatable;
import org.eclipse.e4.core.di.annotations.Optional;
import org.eclipse.e4.core.services.events.IEventBroker;
import org.eclipse.e4.core.services.nls.Translation;
import org.eclipse.e4.ui.di.UIEventTopic;
import org.eclipse.e4.ui.model.application.MApplication;
import org.eclipse.e4.ui.model.application.ui.advanced.MPerspective;
import org.eclipse.e4.ui.model.application.ui.basic.MTrimmedWindow;
import org.eclipse.e4.ui.model.application.ui.basic.MWindow;
import org.eclipse.e4.ui.workbench.modeling.EModelService;
import org.eclipse.e4.ui.workbench.modeling.EPartService;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.widgets.Shell;
import org.osgi.service.prefs.BackingStoreException;
import org.osgi.service.prefs.Preferences;

import de.credo.lookup2.parsers.api.L2ParserConfig;
import de.credo.lookup2.parsers.api.objects.L2DOMObject;
import de.credo.lookup2.parsers.api.structure.L2ParentElement;
import de.credo.lookup2.parsers.impl.RootElement;
import de.credo.lookup2.plugins.api.L2Config;
import de.credo.lookup2.plugins.api.L2Config.L2ConfigType;
import de.credo.lookup2.plugins.api.L2Constants;
import de.credo.lookup2.plugins.api.L2FileData;
import de.credo.lookup2.plugins.api.L2Format;
import de.credo.lookup2.plugins.api.L2ProgressMonitor;
import de.credo.lookup2.plugins.impl.FileData;
import de.credo.lookup2.plugins.impl.ModulesHandler;
import de.credo.lookup2.plugins.impl.PerspectiveTypes;
import de.credo.lookup2.plugins.impl.prefpages.FileMappingsPreferencePage;
import de.credo.lookup2.prefpages.MyPrefPageHandler;
import de.credo.lookup2.viewers.api.L2ViewerConfig;
import jakarta.inject.Inject;
import jakarta.inject.Singleton;

@Creatable @Singleton
public class GlobalAppManager {

	private	final static	int		MAX_RECENT_FILES		= 12;
	private	final static	String	RECENT_FILES_NODE		= "RecentFiles";
	private	final static	String	TWO_STEP_CLOSE_NODE		= "TwoStepClose";
	private	final static	String	START_PERSPECTIVE_NODE	= "StartPerspective";

	private	final static	IEclipsePreferences	userPreferences	= InstanceScope.INSTANCE.getNode(GlobalAppManager.class.getName());

	@Optional
	@Inject	Shell				shell;
	@Inject	MApplication		application;
	@Inject	EModelService		modelService;
	@Inject	IEventBroker		eventBroker;
	@Inject	ModulesHandler		modulesHandler;
	@Inject	MyPrefPageHandler	prefPageHandler;
	@Inject	@Translation	Messages	messages;

	private	L2FileData	lastProcessedFileData;

	public void openInNewWindow(Path file, boolean writable, boolean checkAlreadyOpen, SwitchPerspectiveTargetTypes switchPerspectiveTarget) {
		openInNewWindow(new FileData(file, writable), checkAlreadyOpen, switchPerspectiveTarget);
	}

	public void openInNewWindow(L2FileData fileData, boolean checkAlreadyOpen, SwitchPerspectiveTargetTypes switchPerspectiveTarget) {
		if (checkAlreadyOpen) {
			for (MWindow window: application.getChildren()) {
				L2FileData	openFile	= window.getContext().get(L2FileData.class);
				if (openFile != null && openFile.getFile().equals(fileData.getFile())) {
					// Datei bereits in einem Fenster geöffnet. Dieses nach vorne holen!
					application.setSelectedElement(window);
					return;
				}
			}
		}

		// Datei noch nicht geöffnet, neues Fenster aufmachen, dabei gleichzeitig initial dessen Startperspektive setzen
		MTrimmedWindow	newWindow	= (MTrimmedWindow) modelService.cloneSnippet(application, Constants.MAIN_WINDOW, null);
		application.getChildren().add(newWindow);

		// Übergeben der zu öffnen Datei in den Kontext des neuen Fensters
		openInWindow(newWindow, fileData, switchPerspectiveTarget);
	}

	@SuppressWarnings("incomplete-switch")
	public void openInWindow(MTrimmedWindow window, L2FileData fileData, SwitchPerspectiveTargetTypes switchPerspectiveTarget) {
		// Nachrichtenfenster leeren
		eventBroker.send(Constants.CLEAR_MESSAGES_EVENT, window);

		L2ParserConfig	parserConfig	= null;
		L2ViewerConfig	viewerConfig	= null;
		RootElement		rootElement		= null;

		if (fileData.getFormat() != null) {
			parserConfig	= (L2ParserConfig) modulesHandler.getParserConfig(fileData.getFormat());
			viewerConfig	= (L2ViewerConfig) modulesHandler.getViewerConfig(fileData.getFormat());

		} else {
			List<Map<L2ConfigType, L2Config>>	configSets	= modulesHandler.getConfigSets(fileData.getFile());

			if (configSets.isEmpty()) {
				// Kein vordefiniertes ConfigSet zu diesem Dateinamen gefunden. Rückfrage! Der Anwender darf hier aus einer kompletten
				// Liste von Formaten auswählen, von denen ein möglicherweise eingestelltes Default-Format vorausgewählt wird.
				List<L2Format>	formats	= modulesHandler.getFormats().stream()
						.filter(f -> modulesHandler.creatable(f))
						.sorted()
						.collect(Collectors.toList());
				FormatSelector	dialog	= new FormatSelector(shell, formats, modulesHandler.getDefaultFormat());
				if (dialog.open() == Window.OK) {
					// Der künstliche erste Eintrag zum Hinzufügen zur Konfiguration muss separat abgehandelt werden.
					if (dialog.getSelection() != null) {
						parserConfig	= (L2ParserConfig) modulesHandler.getParserConfig(dialog.getSelection());
						viewerConfig	= (L2ViewerConfig) modulesHandler.getViewerConfig(dialog.getSelection());
					} else {
						prefPageHandler.openPreferencePage(FileMappingsPreferencePage.class, (Shell) window.getWidget());
					}
				}

			} else if (configSets.size() > 1) {
				// Zu viele ConfigSets passen auf dieses Dateinamensmuster. Auch doof. Rückfrage! Der Anwender muss hier aus der Liste
				// der in Frage kommenden Formate einen Typ auswählen, von denen das erste gefundene Format vorausgewählt wird.
				List<L2Format>	formats	= new ArrayList<>();
				for (Map<L2ConfigType, L2Config> cs: configSets) {
					for (L2Config c: cs.values()) {
						if (!formats.contains(c.getFormat())) {
							formats.add(c.getFormat());
						}
					}
				}
				Collections.sort(formats);

				FormatSelector	dialog	= new FormatSelector(shell, formats, formats.get(0));
				if (dialog.open() == Window.OK) {
					// Der künstliche erste Eintrag zum Hinzufügen zur Konfiguration muss separat abgehandelt werden.
					if (dialog.getSelection() != null) {
						parserConfig	= (L2ParserConfig) modulesHandler.getParserConfig(dialog.getSelection());
						viewerConfig	= (L2ViewerConfig) modulesHandler.getViewerConfig(dialog.getSelection());
					} else {
						prefPageHandler.openPreferencePage(FileMappingsPreferencePage.class, (Shell) window.getWidget());
					}
				}

			} else {
				// So wünscht man sich das. Keine Rückfrage
				parserConfig	= (L2ParserConfig) configSets.get(0).get(L2ConfigType.PARSER);
				viewerConfig	= (L2ViewerConfig) configSets.get(0).get(L2ConfigType.VIEWER);
			}
		}

		if (viewerConfig != null || parserConfig != null) {
			window.setLabel("Lookup2" + (fileData != null && fileData.getFile() != null ? " - " + fileData.toString() : ""));

			window.getContext().set(L2FileData.class, fileData);

			window.getContext().set(L2ViewerConfig.class, viewerConfig);

			if (parserConfig != null) {
				window.getContext().set(L2ParserConfig.class, parserConfig);

				if (parserConfig.hasObjects()) {
					modelService.find(Constants.OBJECTS_VIEW_PARTSTACK, window).setToBeRendered(viewerConfig != null);
				}
				if (parserConfig.hasStructure()) {
					modelService.find(Constants.STRUCTURE_VIEW_PARTSTACK, window).setToBeRendered(viewerConfig != null);
				}

				// StructurePart und ObjectPart kriegen ein einzelnes virtuelles RootElement wenn möglich, denn hier bauen ja
				// zwei Perspektiven auf diesem einem gemeinsamen Parser auf, der von diesem RootElement direkt abgeleitet ist.
				// Nur, wenn die Datei grundsätzlich geparst werden kann, wird also in beiden Fenstern überhaupt etwas angezeigt.
				// Sollte, warum auch immer ein null-Wert zurückkommen, dann wird das möglicherweise aus einem vorherigen Lauf
				// noch vorhandene RootElement absichtlich durch das Inject eines null-Values gelöscht.
				rootElement	= createRootElement(window, fileData, parserConfig);
				window.getContext().set(RootElement.class, rootElement);

				// Frühzeitiges und damit möglichst umfassendes Anlegen der Parsing-History
				window.getContext().set(History.class, ContextInjectionFactory.make(History.class, window.getContext()));
			}
		}

		if (switchPerspectiveTarget != SwitchPerspectiveTargetTypes.NONE) {
			PerspectiveTypes	perspectiveType	= getStartPerspective();

			switch (switchPerspectiveTarget) {
			case PARSER:
				if (parserConfig != null) {
					perspectiveType	= parserConfig.hasObjects() ? PerspectiveTypes.OBJECTS : PerspectiveTypes.STRUCTURE;
				}
				break;

			case VIEWER:
				if (viewerConfig != null) {
					perspectiveType	= PerspectiveTypes.VIEWER;
				}
				break;

			case ANY:
				if (parserConfig != null) {
					perspectiveType	= parserConfig.hasObjects() ? PerspectiveTypes.OBJECTS : PerspectiveTypes.STRUCTURE;
				} else if (viewerConfig != null) {
					perspectiveType	= PerspectiveTypes.VIEWER;
				}
				break;
			}

			switchPerspective(window, perspectiveType);

		}
	}

	public void closeFile(MTrimmedWindow window, boolean switchPerspective) {
		window.getContext().get(L2ProgressMonitor.class).setCanceled(true);

		if (window.getContext().get(RootElement.class) != null) {
			window.getContext().get(RootElement.class).closeStream();
		}

		// Zum Glück signalisiert auch das Löschen von Objekten aus dem Kontext den anderen Parts, dass hier was passiert ist
		window.getContext().remove(L2FileData.class);
		window.getContext().remove(L2ViewerConfig.class);
		window.getContext().remove(L2ParserConfig.class);
		window.getContext().remove(RootElement.class);
		window.getContext().remove(L2ParentElement.class);
		window.getContext().remove(L2DOMObject.class);
		window.setLabel("Lookup2");
		eventBroker.send(Constants.CLEAR_MESSAGES_EVENT, window);

		if (switchPerspective) {
			switchPerspective(window, getStartPerspective());
		}
	}

	public void switchPerspective(MTrimmedWindow window, String perspectiveTypeTag) {
		switchPerspective(window, PerspectiveTypes.from(perspectiveTypeTag));
	}

	public void switchPerspective(MTrimmedWindow window, PerspectiveTypes perspectiveType) {
		EPartService	partService		= window.getContext().get(EPartService.class);
		String			perspectiveId	= determinePerspectiveId(window, perspectiveType);
		partService.switchPerspective(perspectiveId);
	}

	public String determinePerspectiveId(MTrimmedWindow window, PerspectiveTypes perspectiveType) {
		return modelService.findElements(window, null, MPerspective.class, Arrays.asList(perspectiveType.getTypeTag())).get(0).getElementId();
	}

	@Inject	@Optional
	public void switchToObjectsPerspective(@UIEventTopic(L2Constants.SWITCH_TO_OBJECTS_PERSPECTIVE_EVENT) MTrimmedWindow desiredWindow, @Active MTrimmedWindow window) {
		if (window != null && window.equals(desiredWindow)) {
			switchPerspective(window, PerspectiveTypes.OBJECTS);
		}
	}

	@Inject	@Optional
	public void switchToStructurePerspective(@UIEventTopic(L2Constants.SWITCH_TO_STRUCTURE_PERSPECTIVE_EVENT) MTrimmedWindow desiredWindow, @Active MTrimmedWindow window) {
		if (window != null && window.equals(desiredWindow)) {
			switchPerspective(window, PerspectiveTypes.STRUCTURE);
		}
	}

	@Inject	@Optional
	public void openLogFile(@UIEventTopic(Constants.OPEN_LOG_FILE_EVENT) Map<Class<?>, Object> values, @Active MTrimmedWindow window) {
		if (values != null) {
			Path			logFile			= (Path) values.get(Path.class);
			MTrimmedWindow	desiredWindow	= (MTrimmedWindow)	values.get(MTrimmedWindow.class);

			// Dieses Event wird für alle Fenster gefeuert, obwohl die Aktion nur für ein spezielles gedacht ist
			if (window.equals(desiredWindow)) {
				openInNewWindow(new FileData(logFile, false), false, SwitchPerspectiveTargetTypes.VIEWER);
			}
		}
	}

	public void storeFileHistory(L2FileData fileData) {
		if (fileData == null || fileData.getFile() == null || fileData == lastProcessedFileData)
			return;

		lastProcessedFileData	= fileData;

		LinkedList<Path>	recentFiles	= loadFileHistory();
		recentFiles.remove(fileData.getFile());		// Irgendwo aus der Mitte löschen, falls die Datei schon einmal aufgerufen worden war
		recentFiles.addFirst(fileData.getFile());	// Vorne als erste Datei neu einfügen
		persistFileHistory(recentFiles);
	}

	public LinkedList<Path> loadFileHistory() {
		Preferences			recentFilesNode	= userPreferences.node(RECENT_FILES_NODE);
		Map<Integer, Path>	recentFiles		= new TreeMap<>();

		try {
			for (String i: recentFilesNode.keys()) {
				String	file	= recentFilesNode.get(i, null);
				Path	path	= Paths.get(file);
				if (Files.exists(path)) {
					recentFiles.put(Integer.parseInt(i), path);
				}
			}
		} catch (BackingStoreException e) {
		}

		return new LinkedList<>(recentFiles.values());
	}

	public void persistFileHistory(List<Path> recentFiles) {
		Preferences	recentFilesNode	= userPreferences.node(RECENT_FILES_NODE);

		try {
			for (String nodeKey: recentFilesNode.childrenNames()) {
				recentFilesNode.remove(nodeKey);
			}

			for (int i = 0; i < recentFiles.size() && i < MAX_RECENT_FILES; i++) {
				recentFilesNode.put(Integer.toString(i), recentFiles.get(i).toAbsolutePath().toString());
			}

			userPreferences.flush();
		} catch (BackingStoreException e) {
		}
	}

	public boolean getTwoStepCloseMode() {
		return userPreferences.getBoolean(TWO_STEP_CLOSE_NODE, true);
	}

	public void setTwoStepCloseMode(boolean twoStepCloseMode) {
		try {
			userPreferences.putBoolean(TWO_STEP_CLOSE_NODE, twoStepCloseMode);
			userPreferences.flush();
		} catch (BackingStoreException e) {
		}
	}

	public PerspectiveTypes getStartPerspective() {
		return PerspectiveTypes.from(userPreferences.get(START_PERSPECTIVE_NODE, PerspectiveTypes.FILES.getTypeTag()));
	}

	public void setStartPerspective(PerspectiveTypes perspectiveType) {
		try {
			userPreferences.put(START_PERSPECTIVE_NODE, perspectiveType.getTypeTag());
			userPreferences.flush();
		} catch (BackingStoreException e) {
		}
	}

	public void dispose() {
		modulesHandler.dispose();
	}

	private RootElement createRootElement(MTrimmedWindow window, L2FileData fileData, L2ParserConfig config) {
		RootElement	rootElement	= ContextInjectionFactory.make(RootElement.class, window.getContext());

		rootElement.openStream(fileData.getFile());

		// Root-Parser erzeugen (oder auch nicht, falls nicht passend)
		if (rootElement.createRootParser(config) == null) {
			rootElement.closeStream();
			return null;
		}

		// Alles ok? Nur dann wird ein voll gültiges RootElement zurückgegeben
		return rootElement;
	}

}
