package de.credo.lookup2;

import java.text.NumberFormat;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.CopyOnWriteArrayList;

import jakarta.inject.Inject;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.jobs.IJobChangeEvent;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.core.runtime.jobs.JobChangeAdapter;
import org.eclipse.e4.core.services.events.IEventBroker;
import org.eclipse.e4.core.services.nls.Translation;
import org.eclipse.e4.ui.di.UISynchronize;
import org.eclipse.e4.ui.workbench.UIEvents;
import org.eclipse.swt.widgets.ProgressBar;

import de.credo.lookup2.plugins.api.L2ProgressMonitor;

public final class GlobalProgressMonitor implements L2ProgressMonitor {

	private class JobProperties {

		private	Job	job;
		private	int	worked;
		private	int	maximum;

		private JobProperties(Job job) {
			this.job	= job;
		}
	}

	private	final static	NumberFormat NF;

	private	final static	int	CHECK_CANCELLATION_DELAY	= 1000;		// in Millisekunden
	private	final static	int	UPDATE_PROGRESS_BAR_DELAY	= 3000;		// in Millisekunden

	@Inject	IEventBroker	eventBroker;
	@Inject	UISynchronize	synchronizer;
	@Inject @Translation	Messages	messages;

	private	Timer				checkForCancellationTimer;
	private	Timer				updateProgressBarTimer;
	private	ProgressBar			progressBar;
	private	List<JobProperties>	activeTasks		= new CopyOnWriteArrayList<>();
	private	boolean				updateNeeded	= true;
	private	boolean				canceled		= false;
	private	boolean				checkCanceled	= false;
	private	JobProperties		lastJob			= new JobProperties(null);	// Dummy

	static {
		NF	= NumberFormat.getPercentInstance();
		NF.setMinimumFractionDigits(1);
	}

	@Override
	public void beginTask(final String name, final int totalWork) {
		synchronizer.syncExec(() -> {
			if (!progressBar.isDisposed()) {
				if (activeTasks.size() == 0) {
					// --- no task is running at the moment ---
					progressBar.setSelection(0);
					progressBar.setMaximum(1);
					progressBar.setToolTipText(null);

				} else {
					// --- other tasks are running ---
					progressBar.setMaximum(activeTasks.stream()
							.mapToInt(at -> {
								if (at.job.getName().equals(name)) {
									at.maximum	= totalWork;
								}
								return at.maximum;})
							.sum() + 1);
					progressBar.setToolTipText(NF.format((double) progressBar.getSelection() / progressBar.getMaximum()));
				}
			}
		});
	}

	@Override
	public void worked(final Job job, final int work) {
		if (lastJob.job != job) {
			for (JobProperties at: activeTasks) {
				if (at.job == job) {
					lastJob	= at;
					break;
				}
			}
		}

		if (work < lastJob.worked || work > lastJob.maximum)
			return;

		lastJob.worked	= work;

		if (updateNeeded) {
			updateNeeded	= false;

			synchronizer.asyncExec(() -> {
				if (!progressBar.isDisposed() ) {
					progressBar.setSelection(activeTasks.stream().mapToInt(at -> at.worked).sum());
					progressBar.setToolTipText(NF.format((double) progressBar.getSelection() / progressBar.getMaximum()));
				}
			});
		}
	}

	@Override
	public void worked(int work) {
		throw new UnsupportedOperationException("use L2ProgressMonitor#worked(Job, int) instead!");
	}

	@Override
	public IProgressMonitor addJob(Job job) {
		if (job != null) {
			setCanceled(false);
			activeTasks.add(new JobProperties(job));

			// Schleife zur regelmäßigen Überprüfung, ob der Job gecancelt wurde
			if (activeTasks.size() == 1) {
				checkForCancellationTimer	= new Timer("Cancellation Timer");
				checkForCancellationTimer.schedule(new TimerTask() {

					@Override
					public void run() {
						checkCanceled	= activeTasks.size() > 0;
					}

				}, CHECK_CANCELLATION_DELAY, CHECK_CANCELLATION_DELAY);

				// Schleife zur regelmäßigen Überprüfung, ob der Job gecancelt wurde
				updateProgressBarTimer	= new Timer("Update ProgressBar Timer");
				updateProgressBarTimer.schedule(new TimerTask() {

					@Override
					public void run() {
						updateNeeded	= true;
					}

				}, UPDATE_PROGRESS_BAR_DELAY, UPDATE_PROGRESS_BAR_DELAY);

				synchronizer.asyncExec(() -> {
					if (progressBar != null && !progressBar.isDisposed()) {
						progressBar.setVisible(true);
					}
				});
			}

			eventBroker.post(UIEvents.REQUEST_ENABLEMENT_UPDATE_TOPIC, UIEvents.ALL_ELEMENT_ID);

			job.addJobChangeListener(new JobChangeAdapter() {

				@Override
				public void done(IJobChangeEvent event) {
					for (JobProperties task: activeTasks) {
						if (task.job == event.getJob()) {
							activeTasks.remove(task);
							break;
						}
					}

					// clean-up
					event.getJob().removeJobChangeListener(this);

					if (activeTasks.size() == 0) {
						checkForCancellationTimer.cancel();
						updateProgressBarTimer.cancel();
						checkCanceled	= false;
						updateNeeded	= true;
					}

					synchronizer.asyncExec(() -> {
						// WICHTIG, weil sonst alle folgenden Jobs ebenfalls abgebrochen werden, obwohl der Auslöser längst nicht mehr wahr ist.
						setCanceled(false);

						if (activeTasks.size() > 0) {
							if (progressBar != null && !progressBar.isDisposed()) {
								progressBar.setSelection(activeTasks.stream().mapToInt(at -> at.worked).sum());
								progressBar.setMaximum(activeTasks.stream().mapToInt(at -> at.maximum).sum() + 1);
								progressBar.setToolTipText(NF.format((double) progressBar.getSelection() / progressBar.getMaximum()));
							}
						} else {
							if (progressBar != null && !progressBar.isDisposed()) {
								progressBar.setSelection(0);
								progressBar.setMaximum(1);
								progressBar.setToolTipText(null);
								progressBar.setVisible(false);
							}
						}
					});

					eventBroker.post(UIEvents.REQUEST_ENABLEMENT_UPDATE_TOPIC, UIEvents.ALL_ELEMENT_ID);
				}
			});
		}

		return this;
	}

	@Override
	public boolean isCanceled() {
		if (checkCanceled) {
			checkCanceled	= false;
			return canceled;
		}

		return false;
	}

	@Override
	public void setCanceled(boolean canceled) {
		if (canceled) {
			for (JobProperties tasks: activeTasks) {
				tasks.job.cancel();
			}
		}
		this.canceled	= activeTasks.size() > 0 ? canceled : false;
	}

	public void setProgressBar(ProgressBar progressBar) {
		this.progressBar = progressBar;
	}

	@Override
	public boolean tasksRunning() {
		return !activeTasks.isEmpty();
	}

	@Override
	public void done() {
	}

	@Override
	public void internalWorked(double work) {
	}

	@Override
	public void setTaskName(String name) {
	}

	@Override
	public void subTask(String name) {
	}

}