package de.credo.lookup2.handlers;

import java.util.Arrays;
import java.util.List;
import java.util.Map;

import jakarta.inject.Inject;
import jakarta.inject.Named;

import org.eclipse.e4.core.contexts.Active;
import org.eclipse.e4.core.di.annotations.CanExecute;
import org.eclipse.e4.core.di.annotations.Execute;
import org.eclipse.e4.core.di.annotations.Optional;
import org.eclipse.e4.ui.di.UIEventTopic;
import org.eclipse.e4.ui.model.application.commands.MParameter;
import org.eclipse.e4.ui.model.application.ui.advanced.MPerspective;
import org.eclipse.e4.ui.model.application.ui.basic.MTrimmedWindow;
import org.eclipse.e4.ui.model.application.ui.menu.MHandledToolItem;
import org.eclipse.e4.ui.workbench.modeling.EModelService;
import org.eclipse.e4.ui.workbench.modeling.EPartService;

import de.credo.lookup2.Constants;
import de.credo.lookup2.ToolbarUtils;
import de.credo.lookup2.parsers.api.L2ParserConfig;
import de.credo.lookup2.parsers.api.objects.L2DOMObject;
import de.credo.lookup2.parsers.api.structure.L2ContextProvider;
import de.credo.lookup2.parsers.api.structure.L2Element;
import de.credo.lookup2.parsers.api.structure.L2ParentElement;
import de.credo.lookup2.parsers.impl.RootElement;
import de.credo.lookup2.plugins.impl.PerspectiveTypes;
import de.credo.lookup2.viewers.api.L2ViewerConfig;

public class PerspectiveSelectionHandler {

	@Execute
	public void execute(@Named(Constants.PERSPECTIVE_BUTTON_ID) String perspectiveId, @Optional L2ParserConfig parserConfig, @Optional L2ViewerConfig viewerConfig, @Optional RootElement rootElement, EModelService modelService, EPartService partService, @Active MTrimmedWindow window, @Optional L2ParentElement parentElement, @Optional L2DOMObject domObject) {
		if (!modelService.getActivePerspective(window).getElementId().equals(perspectiveId)
				&& canExecute(perspectiveId, parserConfig, viewerConfig, rootElement, modelService, window)) {
			partService.switchPerspective(perspectiveId);

			// Die alte Lösung, jede Auswahl in den Objects- und StructureTreeParts prophylaktisch im jeweils
			// jeweils Part gleich mit auszulösen, war insbesondere in Hinblick auf die History ungeschickt.
			// Es reicht doch, die derzeit aktuelle Auswahl genau ein einziges Mal zum Zeitpunkt des Perspek-
			// tivenwechsels in das jeweils andere Format zu übersetzen!
			switch (PerspectiveTypes.from(perspectiveId)) {
			case OBJECTS:
				// Selektion im StructureTree auf die im Objektbaum synchronisieren
				if (parentElement != null) {
					L2DOMObject	object	=	parentElement.getContext().getOfType(L2DOMObject.class, true);

					// Nach oben (wg. traverse = true) nichts gefunden? Dann probieren wir es mal nach unten im Baum
					if (object == null) {
						object	=	findDownInternal(parentElement);
					}

					if (object != null) {
						window.getContext().set(L2DOMObject.class, object);
					}
				}
				break;

			case STRUCTURE:
				// Selektion im Objektbaum auf die zugehörige im StructureTree synchronisieren
				if (domObject != null && domObject.getStartElement() != null) {
					window.getContext().set(L2ParentElement.class, domObject.getStartElement());
				}
				break;

			default:
				break;
			}
		}
	}

	private L2DOMObject findDownInternal(L2ParentElement parentElement) {
		L2DOMObject	object	= null;

		for (L2Element element: parentElement.getChildren()) {
			if (element instanceof L2ParentElement) {
				if (element instanceof L2ContextProvider) {
					object	=	element.getContext().getOfType(L2DOMObject.class);
				}

				if (object == null && ((L2ParentElement) element).hasChildren()) {
					object	= findDownInternal((L2ParentElement) element);				// Rekursion!
				}

				if (object != null) {
					break;
				}
			}
		}

		return object;
	}

	// Wird für jeden Button einzeln durchlaufen, also fünfmal für jede Änderung der Anzeige.
	@CanExecute
	public boolean canExecute(@Named(Constants.PERSPECTIVE_BUTTON_ID) String perspectiveId, @Optional L2ParserConfig parserConfig, @Optional L2ViewerConfig viewerConfig, @Optional RootElement rootElement, EModelService modelService, @Active MTrimmedWindow window) {
		MPerspective		perspectiveForId	= (MPerspective) modelService.find(perspectiveId, window);
		PerspectiveTypes	perspectiveType 	= perspectiveTypeFromTag(perspectiveForId);
		boolean				active				= false;

		switch (perspectiveType) {
		case EMPTY:
			active	= true;
			break;

		case FILES:
			active	= true;
			break;

		case OBJECTS:
			if (rootElement != null && parserConfig != null) {
				active	= parserConfig != null && parserConfig.hasObjects() && parserConfig.creatable();
			}
			break;

		case STRUCTURE:
			if (rootElement != null && parserConfig != null) {
				active	= parserConfig != null && parserConfig.hasStructure() && parserConfig.creatable();
			}
			break;

		case VIEWER:
			if (viewerConfig != null) {
				active	= viewerConfig != null && viewerConfig.creatable();
			}
			break;

		default:
			break;
		}

//		// Notfall-Fallback, falls die aktive Perspektive inaktiv wird -> auf eine aktive umschalten
//		if (!active && activePerspective.getElementId().equals(perspectiveId)) {
//			List<MHandledToolItem>	buttons	= modelService.findElements(window, null, MHandledToolItem.class, Arrays.asList(new String[] {"PerspectiveSelectionButtonGroup"}));
//
//			for (MHandledToolItem b: buttons) {
//				if (b.isEnabled()) {
//					for (MParameter p: b.getParameters()) {
//						if (p.getName().equals(Constants.PERSPECTIVE_BUTTON_ID)) {
//							setSelection(p.getValue(), partService, modelService, window);
//							return true;
//						}
//					}
//				}
//			}
//		}

		return active;
	}

	private PerspectiveTypes perspectiveTypeFromTag(MPerspective perspective) {
		for (String typeTag: perspective.getTags()) {
			for (PerspectiveTypes pt: PerspectiveTypes.values()) {
				if (pt.getTypeTag().equals(typeTag))
					return pt;
			}
		}

		return PerspectiveTypes.EMPTY;
	}

	@Inject
	@Optional
	public void setButton(@UIEventTopic(Constants.SET_PERSPECTIVE_BUTTON_EVENT) Map<Class<?>, Object> values, EModelService modelService, @Active MTrimmedWindow window) {
		if (values != null) {
			String			perspectiveId	= (String) values.get(String.class);
			MTrimmedWindow	desiredWindow	= (MTrimmedWindow)	values.get(MTrimmedWindow.class);

			// Dieses Event wird für alle Fenster gefeuert, obwohl die Aktion nur für ein spezielles gedacht ist
			if (window.equals(desiredWindow)) {
				List<MHandledToolItem>	buttons	= modelService.findElements(window, null, MHandledToolItem.class, Arrays.asList(new String[] {"PerspectiveSelectionButtonGroup"}));
				for (MHandledToolItem button: buttons) {
					for (MParameter parameter: button.getParameters()) {
						if (parameter.getValue().contains(perspectiveId)) {
							button.setSelected(true);
							button.setIconURI(button.getPersistedState().get(Constants.SELECTED_ICON_URI_KEY));
						} else {
							button.setSelected(false);
							button.setIconURI(button.getPersistedState().get(Constants.UNSELECTED_ICON_URI_KEY));
						}
					}
				}

				ToolbarUtils.showPerspectiveToolbars(window, perspectiveId, true);
			}
		}
	}
}