
package de.credo.lookup2.handlers.messages;

import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.time.Instant;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import jakarta.inject.Inject;

import org.eclipse.e4.core.di.annotations.CanExecute;
import org.eclipse.e4.core.di.annotations.Execute;
import org.eclipse.e4.core.di.annotations.Optional;
import org.eclipse.e4.core.services.events.IEventBroker;
import org.eclipse.e4.ui.model.application.ui.basic.MPart;
import org.eclipse.e4.ui.model.application.ui.basic.MTrimmedWindow;

import de.credo.common.bundles.messages.CdoLogMessage;
import de.credo.common.bundles.messages.CdoLogSink;
import de.credo.lookup2.Constants;
import de.credo.lookup2.parts.MessagePart;

public class SaveLogMessagesHandler {

	private	final static	DateTimeFormatter	TIME_FORMATTER	= DateTimeFormatter.ofPattern("HH:mm:ss.SSS").withZone(ZoneId.systemDefault());
	private	final static	DateTimeFormatter	FILE_FORMATTER	= DateTimeFormatter.ofPattern("yyyyMMdd-HHmmss").withZone(ZoneId.systemDefault());

	@Inject	IEventBroker	eventBroker;
	@Optional
	@Inject	CdoLogSink			logSink;		// Wird mit etwas Verzögerung wegen asynchroner Ausführung vom WindowCreationObserver in den Window-Context gesetzt

	private class LogLine {
		String	timestamp	= "";
		String	severity	= "";
		String	origin		= "";
		String	errorcode	= "";
		String	message		= "";

		LogLine(String message) {
			this.message	= message;
		}
	}

	@Execute
	public void execute(MPart part, MTrimmedWindow window) {
		MessagePart	messagePart	= (MessagePart) part.getObject();
		String		logFilename	= "Lookup2_" + FILE_FORMATTER.format(Instant.now()) + ".log";
// Temp-Pfade machen unter Windows Probleme, weil sie in einem versteckten Ordner "AppData" liegen und deswegen das TreePart#reveal den Eintrag
// nur findet, wenn zuvor die Anzeige versteckter Dateien eingeschaltet worden war.
//		Path		logPath		= Paths.get(System.getProperty("java.io.tmpdir"), logFilename);
// Übergangsweise bis zu einer anderen Lösung ins User-Home:
		Path		logPath		= Paths.get(System.getProperty("user.home"), logFilename);

		List<LogLine>	logLines	= new ArrayList<>();
		int	timestampLength	= 12;
		int	severityLength	= 0;
		int	originLength	= 0;
		int	errorcodeLength	= 0;

		// Spaltenbreiten ermitteln
		for (CdoLogMessage message: logSink.getLogMessages()) {
			if (messagePart.getMessageBits().get(message.getSeverity().ordinal())) {
				LogLine	logLine	= new LogLine(message.getMessage());
				logLines.add(logLine);

				logLine.timestamp	= TIME_FORMATTER.format(message.getTimestamp());
				if (logLine.timestamp.length() > timestampLength) {
					severityLength	= logLine.timestamp.length();
				}

				if (message.getSeverity() != null) {
					logLine.severity	= message.getSeverity().name();
					if (logLine.severity.length() > severityLength) {
						severityLength	= logLine.severity.length();
					}
				}

				if (message.getOrigin() != null) {
					logLine.origin		= message.getOrigin();
					if (logLine.origin.length() > originLength) {
						originLength	= logLine.origin.length();
					}
				}

				if (message.getReason() != null) {
					logLine.errorcode	= message.getReason();
					if (logLine.errorcode.length() > errorcodeLength) {
						errorcodeLength	= logLine.errorcode.length();
					}
				}

				if (message.getThrowable() != null) {
					for (StackTraceElement ste: message.getThrowable().getStackTrace()) {
						logLines.add(new LogLine("    " + ste.toString()));
					}
				}
			}
		}

		timestampLength	+= 3;
		severityLength	+= 3;
		originLength	+= 3;
		errorcodeLength	+= 3;

		try (BufferedWriter	br	= Files.newBufferedWriter(logPath, StandardOpenOption.CREATE_NEW, StandardOpenOption.WRITE)) {
			for (LogLine ll: logLines) {
				StringBuilder	sb	= new StringBuilder(ll.timestamp);
				for (int i = ll.timestamp.length(); i < timestampLength; i++) {
					sb.append(' ');
				}

				sb.append(ll.severity);
				for (int i = ll.severity.length(); i < severityLength; i++) {
					sb.append(' ');
				}

				sb.append(ll.origin);
				for (int i = ll.origin.length(); i < originLength; i++) {
					sb.append(' ');
				}

				sb.append(ll.errorcode);
				for (int i = ll.errorcode.length(); i < errorcodeLength; i++) {
					sb.append(' ');
				}

				sb.append(ll.message);
				sb.append(System.lineSeparator());

				br.write(sb.toString());
			}

			br.flush();
			br.close();
		} catch (IOException e) {
		}

		// Weiche Verknüpfung von MessagePart mit FileTreePart. Weil Konstellationen denkbar sind, in denen es zwar einen MessagePart,
		// aber keinen FileTreePart gibt, wurde die Konstante lieber im MessagePart angelegt. Ansonsten gäbe es wieder eine harte
		// Abhängigkeit (was natürlich nicht stimmt, weil ja jederzeit einfach der gleiche String zweimal definiert werden kann)
		Map<Class<?>, Object>	values	= new HashMap<>(3);
		values.put(MTrimmedWindow.class, window);
		values.put(Path.class, logPath);
		eventBroker.post(Constants.OPEN_LOG_FILE_EVENT, values);
	}


	@CanExecute
	public boolean canExecute(MPart part) {
		return logSink != null && !logSink.getLogMessages().isEmpty();
	}

}