package de.credo.lookup2.parts;

import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.BitSet;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

import jakarta.annotation.PostConstruct;
import jakarta.annotation.PreDestroy;
import jakarta.inject.Inject;

import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.preferences.IEclipsePreferences;
import org.eclipse.core.runtime.preferences.InstanceScope;
import org.eclipse.e4.core.contexts.Active;
import org.eclipse.e4.core.di.annotations.Optional;
import org.eclipse.e4.core.services.events.IEventBroker;
import org.eclipse.e4.core.services.nls.Translation;
import org.eclipse.e4.ui.di.Focus;
import org.eclipse.e4.ui.di.UIEventTopic;
import org.eclipse.e4.ui.di.UISynchronize;
import org.eclipse.e4.ui.model.application.ui.basic.MTrimmedWindow;
import org.eclipse.e4.ui.workbench.modeling.EPartService;
import org.eclipse.jface.layout.TableColumnLayout;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.resource.ImageRegistry;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.jface.viewers.CellEditor;
import org.eclipse.jface.viewers.CellLabelProvider;
import org.eclipse.jface.viewers.ColumnViewerToolTipSupport;
import org.eclipse.jface.viewers.ColumnWeightData;
import org.eclipse.jface.viewers.EditingSupport;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.OwnerDrawLabelProvider;
import org.eclipse.jface.viewers.StyledCellLabelProvider;
import org.eclipse.jface.viewers.StyledString;
import org.eclipse.jface.viewers.StyledString.Styler;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.viewers.ViewerCell;
import org.eclipse.jface.viewers.ViewerComparator;
import org.eclipse.swt.SWT;
import org.eclipse.swt.dnd.DND;
import org.eclipse.swt.dnd.DropTarget;
import org.eclipse.swt.dnd.DropTargetEvent;
import org.eclipse.swt.dnd.DropTargetListener;
import org.eclipse.swt.dnd.FileTransfer;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.graphics.TextStyle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.osgi.framework.Bundle;
import org.osgi.framework.FrameworkUtil;
import org.osgi.service.prefs.BackingStoreException;

import de.credo.common.MiscUtils;
import de.credo.common.bundles.messages.CdoLogMessage;
import de.credo.common.bundles.messages.CdoLogSink;
import de.credo.lookup2.Constants;
import de.credo.lookup2.GlobalAppManager;
import de.credo.lookup2.Messages;
import de.credo.lookup2.SwitchPerspectiveTargetTypes;
import de.credo.lookup2.parsers.api.objects.L2DOMObject;
import de.credo.lookup2.parsers.api.structure.L2Element;
import de.credo.lookup2.parsers.api.structure.L2ParentElement;
import de.credo.lookup2.plugins.impl.FileData;
import de.credo.lookup2.plugins.impl.PerspectiveTypes;

public class MessagePart {

	@Inject	Display				display;
	@Inject	MTrimmedWindow		window;
	@Optional
	@Inject	CdoLogSink			logSink;		// Wird mit etwas Verzögerung wegen asynchroner Ausführung vom WindowCreationObserver in den Window-Context gesetzt
	@Inject	EPartService		partService;
	@Inject	GlobalAppManager	globalAppManager;
	@Inject	UISynchronize		syncronizer;
	@Inject	IEventBroker		eventBroker;
	@Inject @Translation	Messages	messages;

	public enum MessageTableColumn {
		TIMESTAMP,
		SEVERITY,
		ORIGIN,
		REASON,
		MESSAGE;
	}

	private	final static	DateTimeFormatter	timeFormatter	= DateTimeFormatter.ofPattern("HH:mm:ss.SSS").withZone(ZoneId.systemDefault());
	private	final static	IEclipsePreferences	userPreferences	= InstanceScope.INSTANCE.getNode(MessagePart.class.getName());

	private	final static	String	MESSAGE_FLAGS_NODE	= "messageFlags";

	private	final static	IPath	VERBOSE_IMAGE_PATH	= new Path("src/main/resources/icons/_16/verbose_icon.png");
	private	final static	IPath	DEBUG_IMAGE_PATH	= new Path("src/main/resources/icons/_16/debug_icon.png");
	private	final static	IPath	INFO_IMAGE_PATH		= new Path("src/main/resources/icons/_16/info_icon.png");
	private	final static	IPath	WARNING_IMAGE_PATH	= new Path("src/main/resources/icons/_16/warning_icon.png");
	private	final static	IPath	ERROR_IMAGE_PATH	= new Path("src/main/resources/icons/_16/error_icon.png");
	private	final static	IPath	FATAL_IMAGE_PATH	= new Path("src/main/resources/icons/_16/fatal_icon.png");

	private	final static	IPath	DROPDOWN_IMAGE_PATH	= new Path("src/main/resources/icons/_16/drop_down_button.png");
	private	final static	String	DROPDOWN_IMAGE_KEY	= "DROP_DOWN_BUTTON_KEY";

	private	final static	int		refreshDelay		= 1000;		// in Millisekunden

	private	Composite			parent;
	private	BitSet				messageBits		= new BitSet(6);	// verbose	= 0, debug	= 1, info	= 2, warning	= 3, error	= 4, fatal	= 5
	private	TableViewer			tableViewer;
	private	MessageComparator	comparator;
	private	ImageRegistry		imageRegistry;
	private	Timer				refreshTimer;

	private	int					oldMessageCount	= -1;

	private class MessageComparator extends ViewerComparator {

		private	MessageTableColumn	tableColumn	= MessageTableColumn.TIMESTAMP;
		private	boolean				ascending	= false;

		@Override
		public int compare(Viewer viewer, Object e1, Object e2) {
			CdoLogMessage	l1	= (CdoLogMessage) e1;
			CdoLogMessage	l2	= (CdoLogMessage) e2;

			switch (tableColumn) {
			case TIMESTAMP:	return (ascending ? 1 : -1) * l1.getTimestamp().compareTo(l2.getTimestamp());
			case SEVERITY:	return (ascending ? 1 : -1) * l1.getSeverity().compareTo(l2.getSeverity());
			case ORIGIN:	return (ascending ? 1 : -1) * l1.getOrigin().compareTo(l2.getOrigin());
			case REASON:	return (ascending ? 1 : -1) * l1.getReason().compareTo(l2.getReason());
			case MESSAGE:	return (ascending ? 1 : -1) * l1.getMessage().compareTo(l2.getMessage());
			}

			return super.compare(viewer, e1, e2);
		}

		public MessageTableColumn getTableColumn() {
			return tableColumn;
		}

		public void setTableColumn(MessageTableColumn tableColumn) {
			this.tableColumn	= tableColumn;
		}

		public boolean isAscending() {
			return ascending;
		}

		public void setAscending(boolean ascending) {
			this.ascending	= ascending;
		}
	}

	private static Styler LINK_STYLER	= new Styler() {
		@Override
		public void applyStyles(TextStyle textStyle) {
//			textStyle.foreground		= Display.getCurrent().getSystemColor(SWT.COLOR_BLUE);
			textStyle.underline			= true;
			textStyle.underlineStyle	= SWT.UNDERLINE_LINK;
		}
	};

	@PostConstruct
	public void postConstruct(Composite parent) {
		this.parent	= parent;

		parent.setLayout(new FillLayout());

		imageRegistry	= JFaceResources.getImageRegistry();

		createTable(parent);

		// Laden der Voreinstellungen und Weitergeben an die Toolbar
		loadPreferences();
		eventBroker.post(Constants.ACTIVE_MESSAGE_LEVELS_EVENT, messageBits);

		if (logSink != null) {
			logSinkInjector(logSink);
		}
	}

	private void createTable(Composite parent) {
		Composite	tableComposite	= new Composite(parent, SWT.NONE);
		tableComposite.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));

		TableColumnLayout	tableLayout	= new TableColumnLayout();
		tableComposite.setLayout(tableLayout);

		tableViewer	= new TableViewer(tableComposite, SWT.MULTI | SWT.FULL_SELECTION);
		tableViewer.getTable().setHeaderVisible(true);
		tableViewer.getTable().setLinesVisible(true);

		ColumnViewerToolTipSupport.enableFor(tableViewer);

		tableViewer.setContentProvider(new IStructuredContentProvider() {
			@Override
			public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
				IStructuredContentProvider.super.inputChanged(viewer, oldInput, newInput);
			}

			@SuppressWarnings("unchecked")
			@Override
			public Object[] getElements(Object inputElement) {
				return ((List<CdoLogMessage>) inputElement).stream()
							.filter(m -> messageBits.get(m.getSeverity().ordinal()))
							.toArray();
			}
		});

		comparator	= new MessageComparator();
		tableViewer.setComparator(comparator);

		TableViewerColumn	timeColumn	= new TableViewerColumn(tableViewer, SWT.LEFT);
		timeColumn.getColumn().setText(messages.messagePartTableViewerTimeColumLabel);
		tableLayout.setColumnData(timeColumn.getColumn(), new ColumnWeightData(60, 70, false));			// 6 %
		timeColumn.getColumn().addSelectionListener(createColumnSelectionListener(timeColumn.getColumn(), MessageTableColumn.TIMESTAMP));
		tableViewer.getTable().setSortColumn(timeColumn.getColumn());
		tableViewer.getTable().setSortDirection(comparator.isAscending() ? SWT.DOWN : SWT.UP);
		timeColumn.setLabelProvider(new CellLabelProvider() {
			@Override
			public void update(ViewerCell cell) {
				cell.setText(timeFormatter.format(((CdoLogMessage) cell.getElement()).getTimestamp()));
			}
		});

		TableViewerColumn severityColumn	= new TableViewerColumn(tableViewer, SWT.CENTER);
// Zu wenig Platz für eine sinnvolle Überschrift
//		severityColumn.getColumn().setText(messages.messePartTableViewerSeverityColumLabel);
		severityColumn.getColumn().addSelectionListener(createColumnSelectionListener(severityColumn.getColumn(), MessageTableColumn.SEVERITY));
		tableLayout.setColumnData(severityColumn.getColumn(), new ColumnWeightData(20, 40, false));		// 2 %

		// Zentrieren des Images in der Spalte geht nur über die eigene Kontrolle des Paint-Events.
		// Am Ende passiert über diesen LabelProvider nichts anderes, als das.
		severityColumn.setLabelProvider(new OwnerDrawLabelProvider() {

			@Override
			protected void measure(Event event, Object element) {
			}

			@Override
			protected void paint(Event event, Object element) {
				Image	img	= getImage(element);

				if (img != null) {
					Rectangle	bounds	= ((TableItem) event.item).getBounds(event.index);
					Rectangle	imgBounds	= img.getBounds();
					bounds.width	/= 2;
					bounds.width	-= imgBounds.width / 2;
					bounds.height	/= 2;
					bounds.height	-= imgBounds.height / 2;

					int	x	= bounds.width > 0 ? bounds.x + bounds.width : bounds.x;
					int	y	= bounds.height > 0 ? bounds.y + bounds.height : bounds.y;

					event.gc.drawImage(img, x, y);
				}
			}

			protected Image getImage(Object element) {
				Image	icon	= imageRegistry.get(((CdoLogMessage) element).getSeverity().name());
				if (icon == null) {
					IPath	iconPath	= null;

					switch (((CdoLogMessage) element).getSeverity()) {
					case VERBOSE:	iconPath	= VERBOSE_IMAGE_PATH;		break;
					case DEBUG:		iconPath	= DEBUG_IMAGE_PATH;			break;
					case INFO:		iconPath	= INFO_IMAGE_PATH;			break;
					case WARNING:	iconPath	= WARNING_IMAGE_PATH;		break;
					case ERROR:		iconPath	= ERROR_IMAGE_PATH;			break;
					case FATAL:		iconPath	= FATAL_IMAGE_PATH;			break;
					default:												break;
					}

					Bundle	bundle	= FrameworkUtil.getBundle(getClass());
					URL		imgFile	= FileLocator.find(bundle, iconPath, null);
					icon	= ImageDescriptor.createFromURL(imgFile).createImage();
					imageRegistry.put(((CdoLogMessage) element).getSeverity().name(), icon);
				}

				return icon;
			}
		});

		TableViewerColumn	originColumn	= new TableViewerColumn(tableViewer, SWT.CENTER);
		originColumn.getColumn().setText(messages.messagePartTableViewerOriginColumLabel);
		tableLayout.setColumnData(originColumn.getColumn(), new ColumnWeightData(100, 70, true));		// 10 %
		originColumn.getColumn().addSelectionListener(createColumnSelectionListener(originColumn.getColumn(), MessageTableColumn.ORIGIN));
		originColumn.setLabelProvider(new CellLabelProvider() {

			@Override
			public void update(ViewerCell cell) {
				cell.setText(((CdoLogMessage) cell.getElement()).getOrigin());
			}
		});

		TableViewerColumn	reasonColumn	= new TableViewerColumn(tableViewer, SWT.CENTER);
		reasonColumn.getColumn().setText(messages.messagePartTableViewerReasonColumLabel);
		tableLayout.setColumnData(reasonColumn.getColumn(), new ColumnWeightData(70, 70, true));		// 7 %
		reasonColumn.getColumn().addSelectionListener(createColumnSelectionListener(reasonColumn.getColumn(), MessageTableColumn.REASON));
		reasonColumn.setLabelProvider(new StyledCellLabelProvider() {

			@Override
			public void update(final ViewerCell cell) {
				if (((CdoLogMessage) cell.getElement()).getReference() != null) {
					StyledString	styledString	= new StyledString(((CdoLogMessage) cell.getElement()).getReason(), LINK_STYLER);
					cell.setText(styledString.getString());
					cell.setStyleRanges(styledString.getStyleRanges());
				} else {
					cell.setText(((CdoLogMessage) cell.getElement()).getReason());
					cell.setStyleRanges(null);
				}
			}
		});

		reasonColumn.setEditingSupport(new EditingSupport(tableViewer) {

			@Override
			protected void setValue(Object element, Object value) {
			}

			@Override
			protected Object getValue(Object element) {
				return null;
			}

			@Override
			protected CellEditor getCellEditor(Object element) {
				return null;
			}

			@Override
			// Da dieses Objekt nicht wirklich editiert werden muss, reicht es, via canEdit zu ermitteln, auf welches Objekt geklickt
			// worden war! ist so viel einfacher, als erst umständlich irgendwo einen MouseListener zu implementieren.
			protected boolean canEdit(Object element) {
				Object	reference	= ((CdoLogMessage) element).getReference();

				if (reference != null) {
					if (reference instanceof L2Element) {
						partService.switchPerspective(PerspectiveTypes.STRUCTURE.getElementId());
						window.getContext().set(L2ParentElement.class, (L2ParentElement) ((L2Element) reference instanceof L2ParentElement ? reference : ((L2Element) reference).getTreeParent()));

					} else if (reference instanceof L2DOMObject) {
						partService.switchPerspective(PerspectiveTypes.OBJECTS.getElementId());
						window.getContext().set(L2DOMObject.class, (L2DOMObject) reference);
					}
				}

				return false;
			}
		});

		TableViewerColumn	messageColumn	= new TableViewerColumn(tableViewer, SWT.LEFT);
		messageColumn.getColumn().setText(messages.messagePartTableViewerMessageColumLabel);
		tableLayout.setColumnData(messageColumn.getColumn(), new ColumnWeightData(750, 300, true));		// 75 %
		messageColumn.getColumn().addSelectionListener(createColumnSelectionListener(messageColumn.getColumn(), MessageTableColumn.MESSAGE));
		messageColumn.setLabelProvider(new CellLabelProvider() {

			@Override
			public String getToolTipText(Object element) {
				// Wichtig: Vorherige Definition des ColumnViewerToolTipSupport.enableFor(tableViewer);
				if (((CdoLogMessage) element).getThrowable() != null)
					return MiscUtils.printThrowable(((CdoLogMessage) element).getThrowable());

				return super.getToolTipText(element);
			}

			@Override
			public void update(ViewerCell cell) {
				cell.setText(((CdoLogMessage) cell.getElement()).getMessage());

				if (((CdoLogMessage) cell.getElement()).getThrowable() != null) {
					Image	dropDown	= imageRegistry.get(DROPDOWN_IMAGE_KEY);
					if (dropDown == null) {
						Bundle	bundle	= FrameworkUtil.getBundle(getClass());
						URL		imgFile	= FileLocator.find(bundle, DROPDOWN_IMAGE_PATH, null);
						dropDown		= ImageDescriptor.createFromURL(imgFile).createImage();
						imageRegistry.put(DROPDOWN_IMAGE_KEY, dropDown);
					}
					cell.setImage(dropDown);
				}
			}
		});

		// Schleife zum regelmäßigen Update der Tabelle
		refreshTimer	= new Timer("Refresh MessagePart periodically");
		refreshTimer.schedule(new TimerTask() {

			@Override
			public void run() {
				// Refresh des TableViewers, wenn es Updates gegeben hat
				if (logSink != null && oldMessageCount != logSink.getLogMessages().size()) {
					oldMessageCount	= logSink.getLogMessages().size();

					syncronizer.asyncExec(() -> {
						if (!tableViewer.getControl().isDisposed()) {
							tableViewer.refresh();
						}
					});
				}
			}
		}, refreshDelay, refreshDelay);

		DropTarget	target	= new DropTarget(tableComposite, DND.DROP_COPY | DND.DROP_LINK | DND.DROP_MOVE);
		target.setTransfer(new Transfer[] {FileTransfer.getInstance()});
		target.addDropListener(new DropTargetListener() {

			@Override	public void dropAccept(DropTargetEvent event) {}

			@Override
			public void drop(DropTargetEvent event) {
				if (event.data instanceof String[]) {
					String[]	filenames	= (String[]) event.data;
					if (filenames.length > 0) {
						for (int i	= 0; i < filenames.length; i++) {
							java.nio.file.Path	filePath	= Paths.get(filenames[i]);
							globalAppManager.openInNewWindow(new FileData(filePath, Files.isWritable(filePath)), true, SwitchPerspectiveTargetTypes.ANY);
						}
					}
				}
			}

			@Override	public void dragOver(DropTargetEvent event) {}

			@Override	public void dragOperationChanged(DropTargetEvent event) {}

			@Override	public void dragLeave(DropTargetEvent event) {}

			@Override	public void dragEnter(DropTargetEvent event) {}
		});
	}

	private SelectionListener createColumnSelectionListener(final TableColumn column, MessageTableColumn tableColumn) {
		return new SelectionListener() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				if (comparator.getTableColumn() == tableColumn) {
					comparator.setAscending(!comparator.isAscending());
				} else {
					comparator.setTableColumn(tableColumn);
					comparator.setAscending(true);
				}

				tableViewer.getTable().setSortColumn(column);
				tableViewer.getTable().setSortDirection(comparator.isAscending() ? SWT.DOWN : SWT.UP);
				tableViewer.refresh();
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent e) {
				widgetSelected(e);
			}
		};
	}

	@PreDestroy
	public void dispose() {
		refreshTimer.cancel();
		persistPreferences();
	}

	private void loadPreferences() {
		messageBits	= BitSet.valueOf(new long[] {userPreferences.getLong(MESSAGE_FLAGS_NODE, 0x3C)});		// 00111100 (verbose/debug	= off, alle anderen Flags an)
	}

	private void persistPreferences() {
		try {
			userPreferences.putLong(MESSAGE_FLAGS_NODE, messageBits.toLongArray()[0]);
			userPreferences.flush();
		} catch (BackingStoreException e) {
		}
	}

	@Inject
	@Optional
	public void clearMessages(@UIEventTopic(Constants.CLEAR_MESSAGES_EVENT) MTrimmedWindow window, @Active MTrimmedWindow activeWindow) {
		if (window != null && window == activeWindow) {
			logSink.clearMessages();
			tableViewer.refresh();
		}
	}

	@Focus
	public void setFocus() {
		tableViewer.getTable().setFocus();
	}

	public BitSet getMessageBits() {
		return messageBits;
	}

	public void setMessageBits(BitSet messageBits) {
		this.messageBits	= messageBits;
		tableViewer.refresh();
	}

	@Inject
	public void logSinkInjector(@Optional CdoLogSink logSink) {
		if (this.parent == null)
			return;

		if (logSink != null) {
			tableViewer.setInput(logSink.getLogMessages());
		}
	}

}
