package de.credo.lookup2.parts.files;

import java.io.IOException;
import java.net.URL;
import java.nio.file.Path;
import java.text.NumberFormat;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.time.format.FormatStyle;
import java.util.AbstractMap.SimpleEntry;
import java.util.Arrays;
import java.util.Map.Entry;

import jakarta.annotation.PostConstruct;
import jakarta.inject.Inject;

import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.ICoreRunnable;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.e4.core.contexts.ContextInjectionFactory;
import org.eclipse.e4.core.contexts.IEclipseContext;
import org.eclipse.e4.core.di.annotations.Optional;
import org.eclipse.e4.core.services.events.IEventBroker;
import org.eclipse.e4.core.services.nls.Translation;
import org.eclipse.e4.ui.di.Focus;
import org.eclipse.e4.ui.di.UIEventTopic;
import org.eclipse.e4.ui.di.UISynchronize;
import org.eclipse.e4.ui.model.application.ui.basic.MTrimmedWindow;
import org.eclipse.e4.ui.services.EMenuService;
import org.eclipse.e4.ui.workbench.modeling.EPartService;
import org.eclipse.e4.ui.workbench.modeling.ESelectionService;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.resource.ImageRegistry;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.jface.viewers.ILazyTreeContentProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.ITreeViewerListener;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TreeExpansionEvent;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.TreeViewerColumn;
import org.eclipse.jface.viewers.ViewerCell;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.BusyIndicator;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.ControlListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.program.Program;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.TreeColumn;
import org.eclipse.swt.widgets.TreeItem;
import org.osgi.framework.Bundle;
import org.osgi.framework.FrameworkUtil;

import de.credo.lookup2.Constants;
import de.credo.lookup2.GlobalAppManager;
import de.credo.lookup2.Messages;
import de.credo.lookup2.SwitchPerspectiveTargetTypes;
import de.credo.lookup2.filetree.api.FileTree;
import de.credo.lookup2.filetree.api.FileTreeCategory;
import de.credo.lookup2.filetree.api.FileTreeColumn;
import de.credo.lookup2.filetree.api.FileTreeConstants;
import de.credo.lookup2.filetree.api.FileTreeNode;
import de.credo.lookup2.filetree.api.FileTreeNodeFile;
import de.credo.lookup2.filetree.api.FileTreeNodeFileSystem;
import de.credo.lookup2.plugins.api.L2FileData;
import de.credo.lookup2.plugins.impl.FileData;
import de.credo.lookup2.plugins.impl.ModulesHandler;

public class FileTreePart {

	private final static	String	MAGAZINE_FILLED_IMAGE_KEY	= "Magazine_filled_int";
	private final static	IPath	MAGAZINE_FILLED_IMAGE_PATH	= new org.eclipse.core.runtime.Path("src/main/resources/icons/_16/magazine_folder.png");
	private final static	String	MAGAZINE_EMPTY_IMAGE_KEY	= "Magazine_empty_int";
	private final static	IPath	MAGAZINE_EMPTY_IMAGE_PATH	= new org.eclipse.core.runtime.Path("src/main/resources/icons/_16/magazine_folder_empty.png");
	private final static	String	SERVER_IMAGE_KEY			= "Server_int";
	private final static	IPath	SERVER_IMAGE_PATH			= new org.eclipse.core.runtime.Path("src/main/resources/icons/_16/server_empty.png");
	private final static	String	SERVER_ATTACHED_IMAGE_KEY	= "Server_attached_int";
	private final static	IPath	SERVER_ATTACHED_IMAGE_PATH	= new org.eclipse.core.runtime.Path("src/main/resources/icons/_16/server.png");
	private final static	String	FOLDER_CLOSED_IMAGE_KEY		= "Folder_closed_int";
	private final static	IPath	FOLDER_CLOSED_IMAGE_PATH	= new org.eclipse.core.runtime.Path("src/main/resources/icons/_16/folder.png");
	private final static	String	FOLDER_OPEN_IMAGE_KEY		= "Folder_open_int";
	private final static	IPath	FOLDER_OPEN_IMAGE_PATH		= new org.eclipse.core.runtime.Path("src/main/resources/icons/_16/folder_open.png");

	private final static	int		OPEN_FILE_JOB_DELAY			= 500;		// 500 Millisekunden

	@Inject Display				display;
	@Inject Shell				shell;
	@Inject	IEclipseContext		context;
	@Inject	ESelectionService	selectionService;
	@Inject	UISynchronize		synchronizer;
	@Inject	IEventBroker		eventBroker;
	@Inject	EMenuService		menuService;
	@Inject	GlobalAppManager	globalAppManager;
	@Inject	FileTree			fileTree;
	@Inject	ModulesHandler		modulesHandler;
	@Inject	MTrimmedWindow		window;
	@Inject	EPartService		partService;
	@Inject		@Translation	Messages	messages;

	private	ImageRegistry	imageRegistry;
	private	TreeViewer		treeViewer;
	private Combo			quickSelectCombo;
	private	FileTreeNode	lastSelection;
	private	boolean			ignoreSelectionEvent;
	private	Entry<FileTreeNode, OpenFileActionTypes> fileToOpen;

	@PostConstruct
	public void postConstruct(Composite parent) {
		// Wurde dieser Part schon aktiviert durch das Injecten in eine der Inject-Methoden unten? Gut! -> Zurück
		if (treeViewer != null)
			return;

		// Manuelles Initialisieren der FileTree-Componente. Warum das @PostConstruct nicht automatisch beim Erzeugen der
		// Komponente ausgeführt wird, lies sich auf die Schnelle nicht herausfinden. So geht es auch.
		ContextInjectionFactory.inject(fileTree, context);

		createControls(parent);

		// WICHTIG! Ohne diesen hier generierten Link zwischen Tree und Popup-Menü im Model wird einfach gar nichts angezeigt!
		// https://stackoverflow.com/questions/70986667/creating-a-popup-menu-within-a-fragment-e4xmi-for-an-editor-at-extension-point-o
		createContextMenu();

		eventBroker.post(FileTreeConstants.SET_SHOWHIDDEN_STATE, fileTree.isShowHiddenFiles());
		eventBroker.post(FileTreeConstants.SET_SHOWSUPPORTED_STATE, fileTree.isShowSupportedOnly());

		treeViewer.setInput(fileTree.getRootNode());
	}

	protected void createControls(Composite parent) {
		// Anlegen der Controls
		imageRegistry	= 	JFaceResources.getImageRegistry();

		Composite	composite	= new Composite(parent, SWT.NONE);
		GridLayout	layout		= new GridLayout();
		layout.marginHeight		= 0;
		layout.marginWidth		= 0;
		composite.setLayout(layout);

		if (fileTree.isUseQuickSelection()) {
			createQuickSelectionCombo(composite);
		}

		createTreeControl(composite);
	}

	private void createQuickSelectionCombo(Composite composite) {
		quickSelectCombo = new Combo(composite, SWT.DROP_DOWN);
		quickSelectCombo.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		quickSelectCombo.setFont(composite.getFont());
		quickSelectCombo.setEnabled(true);
		quickSelectCombo.setVisibleItemCount(10);

		quickSelectCombo.addSelectionListener(new SelectionAdapter() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				int	selectionIndex	= ((Combo) e.getSource()).getSelectionIndex();
				if (selectionIndex < 0 && !quickSelectCombo.getText().trim().isEmpty()) {
					FileTreeNode filePath;
					try {
						filePath = fileTree.getNode(quickSelectCombo.getText().trim());
						if (filePath.exists()) {
							revealElement(filePath, false);
						}
					} catch (IOException e1) {
					}
				} else {
					try {
						FileTreeNode	node	= fileTree.getNode(((Combo) e.getSource()).getItem(selectionIndex));
						treeViewer.expandToLevel(node, 1);
						treeViewer.setSelection(new StructuredSelection(node), true);
					} catch (IOException e1) {
					}
				}
				autoSizeColumns();
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent e) {
				widgetSelected(e);
			}
		});

		if (fileTree.getQuickSelectionList() != null && !fileTree.getQuickSelectionList().isEmpty()) {
			quickSelectCombo.setItems(fileTree.getQuickSelectionList().toArray(new String[0]));
		}
	}

	private void createTreeControl(Composite composite) {
		treeViewer = new TreeViewer(composite, SWT.MULTI | SWT.H_SCROLL | SWT.V_SCROLL | SWT.FULL_SELECTION | SWT.VIRTUAL);

		treeViewer.getTree().setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
		treeViewer.getTree().setHeaderVisible(true);
		treeViewer.setUseHashlookup(true);
		treeViewer.setContentProvider(new ILazyTreeContentProvider() {

			@Override
			public Object getParent(Object element) {
				return ((FileTreeNode) element).getParent();
			}

			@Override
			public void updateElement(Object parent, int index) {
				FileTreeNode	node	= ((FileTreeNode) parent).getFilteredAndSorted().get(index);
				node.refresh();
				treeViewer.replace(parent, index, node);
				treeViewer.setHasChildren(node, node.hasChildren());
			}

			@Override
			public void updateChildCount(Object element, int currentChildCount) {
				((FileTreeNode) element).refresh();
				int	childCount	= ((FileTreeNode) element).getChildCount();
				if (childCount != currentChildCount) {
					treeViewer.setChildCount(element, childCount);
				}
			}
		});

		TreeViewerColumn	nameColumn	= new TreeViewerColumn(treeViewer, SWT.NONE);
		nameColumn.getColumn().setText(messages.filePartTreeViewerNameColumnLabel);
		nameColumn.getColumn().setWidth(fileTree.getColumnWidths().get(FileTreeColumn.NAME));
		nameColumn.getColumn().addControlListener(createColumnControlListener(FileTreeColumn.NAME));
		nameColumn.getColumn().addSelectionListener(createColumnSelectionListener(FileTreeColumn.NAME, true));
		nameColumn.setLabelProvider(new ColumnLabelProvider() {

			@Override
			public String getText(Object element) {
				return ((FileTreeNode) element).getName();
			}

			@Override
			public void update(ViewerCell cell) {
				super.update(cell);

				FileTreeNode	element	= (FileTreeNode) cell.getElement();

				Image	icon	= null;

				boolean	expanded	= treeViewer.getExpandedState(element);

				if (element instanceof FileTreeCategory) {
					icon		= imageRegistry.get(expanded ? MAGAZINE_EMPTY_IMAGE_KEY: MAGAZINE_FILLED_IMAGE_KEY);
					if (icon == null) {
						Bundle	bundle	= FrameworkUtil.getBundle(getClass());
						URL		imgFile	= FileLocator.find(bundle, expanded ? MAGAZINE_EMPTY_IMAGE_PATH: MAGAZINE_FILLED_IMAGE_PATH, null);
						icon	= ImageDescriptor.createFromURL(imgFile).createImage();
						imageRegistry.put(expanded ? MAGAZINE_EMPTY_IMAGE_KEY: MAGAZINE_FILLED_IMAGE_KEY, icon);
					}

				} else if (element instanceof FileTreeNodeFileSystem) {
					icon		= imageRegistry.get(expanded ? SERVER_IMAGE_KEY: SERVER_ATTACHED_IMAGE_KEY);
					if (icon == null) {
						Bundle	bundle	= FrameworkUtil.getBundle(getClass());
						URL		imgFile	= FileLocator.find(bundle, expanded ? SERVER_IMAGE_PATH: SERVER_ATTACHED_IMAGE_PATH, null);
						icon	= ImageDescriptor.createFromURL(imgFile).createImage();
						imageRegistry.put(expanded ? SERVER_IMAGE_KEY: SERVER_ATTACHED_IMAGE_KEY, icon);
					}

				} else if (element instanceof FileTreeNodeFile) {
					if (((FileTreeNodeFile) element).isFile()) {
						String ext	= ((FileTreeNodeFile) element).getExtension();
						icon		= imageRegistry.get(ext + "_ext");
						if (icon == null) {
							Program	p	= Program.findProgram(((FileTreeNodeFile) element).getExtension());
							if (p != null) {
								ImageData	iData	= p.getImageData();
								if (iData != null) {
									icon	= new Image(composite.getDisplay(), iData);
									imageRegistry.put(ext + "_ext", icon);
								}
							}
						}

					} else {
						icon	= imageRegistry.get(expanded ? FOLDER_OPEN_IMAGE_KEY : FOLDER_CLOSED_IMAGE_KEY);
						if (icon == null) {
							Bundle	bundle	= FrameworkUtil.getBundle(getClass());
							URL		imgFile	= FileLocator.find(bundle, expanded ? FOLDER_OPEN_IMAGE_PATH : FOLDER_CLOSED_IMAGE_PATH, null);
							icon	= ImageDescriptor.createFromURL(imgFile).createImage();
							imageRegistry.put(expanded ? FOLDER_OPEN_IMAGE_KEY : FOLDER_CLOSED_IMAGE_KEY, icon);
						}
					}
				}

				cell.setImage(icon);
				colorizeCell(cell);
			}
		});

//		TreeViewerColumn	extensionColumn	= new TreeViewerColumn(treeViewer, SWT.CENTER);
//		extensionColumn.getColumn().setText(messages.filePartTreeViewerExtensionColumnLabel);
//		extensionColumn.getColumn().setWidth(fileTree.getColumnWidths().get(FileTreeColumn.EXTENSION));
//		extensionColumn.getColumn().addControlListener(createColumnControlListener(FileTreeColumn.EXTENSION));
//		extensionColumn.getColumn().addSelectionListener(createColumnSelectionListener(FileTreeColumn.EXTENSION));
//		extensionColumn.setLabelProvider(new ColumnLabelProvider() {
//
//			@Override
//			public String getText(Object element) {
//				return ((FileTreeNode) element).getExtension() == null ? "" : ((FileTreeNode) element).getExtension();
//			}
//
//			@Override
//			public void update(ViewerCell cell) {
//				super.update(cell);
//				colorizeCell(cell);
//			}
//		});

		TreeViewerColumn	sizeColumn	= new TreeViewerColumn(treeViewer, SWT.RIGHT);
		sizeColumn.getColumn().setText(messages.filePartTreeViewerSizeColumnLabel);
		sizeColumn.getColumn().setWidth(fileTree.getColumnWidths().get(FileTreeColumn.SIZE));
		sizeColumn.getColumn().addControlListener(createColumnControlListener(FileTreeColumn.SIZE));
		sizeColumn.getColumn().addSelectionListener(createColumnSelectionListener(FileTreeColumn.SIZE, false));
		sizeColumn.setLabelProvider(new ColumnLabelProvider() {

			@Override
			public String getText(Object element) {
				return ((FileTreeNode) element).getSize() < 0 ? "" :
					NumberFormat.getInstance().format((((FileTreeNode) element).getSize() + 1023) / 1024) + " KB";
			}

			@Override
			public void update(ViewerCell cell) {
				super.update(cell);
				colorizeCell(cell);
			}
		});

		TreeViewerColumn	dateColumn	= new TreeViewerColumn(treeViewer, SWT.RIGHT);
		dateColumn.getColumn().setText(messages.filePartTreeViewerDateColumnLabel);
		dateColumn.getColumn().setWidth(fileTree.getColumnWidths().get(FileTreeColumn.DATE));
		dateColumn.getColumn().addControlListener(createColumnControlListener(FileTreeColumn.DATE));
		dateColumn.getColumn().addSelectionListener(createColumnSelectionListener(FileTreeColumn.DATE, false));
		dateColumn.setLabelProvider(new ColumnLabelProvider() {

			@Override
			public String getText(Object element) {
				return ((FileTreeNode) element).getDate() == null ? "" :
					DateTimeFormatter.ofLocalizedDateTime(FormatStyle.SHORT).format(((FileTreeNode) element).getDate().atZone(ZoneId.systemDefault()));
			}

			@Override
			public void update(ViewerCell cell) {
				super.update(cell);
				colorizeCell(cell);
			}
		});

		treeViewer.addTreeListener(new ITreeViewerListener() {

			@Override
			public void treeExpanded(TreeExpansionEvent event) {
				synchronizer.asyncExec(() -> {
					FileTreeNode	node	= (FileTreeNode) event.getElement();
					fileTree.addWatch(node);

					autoSizeColumns();
				});
			}

			@Override
			public void treeCollapsed(TreeExpansionEvent event) {
				synchronizer.asyncExec(() -> {
					FileTreeNode	node	= (FileTreeNode) event.getElement();
					fileTree.removeWatch(node);

					autoSizeColumns();
				});
			}
		});

		treeViewer.getTree().addMouseListener(new MouseListener() {

			@Override
			public void mouseUp(MouseEvent e) {
				if (fileToOpen != null) {
					switch (e.button) {
					case 1:													// Linke Maustaste ...
						switch (e.count) {
						case 1:												// ... und nur beim ersten Klick
							if ((e.stateMask & SWT.CTRL) != SWT.CTRL) {		// ... und nur, wenn nicht Ctrl gehalten wird
								fileToOpen.setValue(OpenFileActionTypes.SINGLE_CLICK);
							} else {
								fileToOpen	= null;
							}
							break;
						}
						break;

					case 3:													// Rechte Maustaste ...
						switch (e.count) {
						case 1:												// ... und nur beim ersten Klick
							if ((e.stateMask & SWT.CTRL) != SWT.CTRL) {		// ... und nur, wenn nicht Ctrl gehalten wird
								fileToOpen.setValue(OpenFileActionTypes.RIGHT_CLICK);
							} else {
								fileToOpen	= null;
							}
							break;
						}
						break;

					default:
						fileToOpen	= null;
					}
				}
			}

			@Override
			public void mouseDown(MouseEvent e) {
				switch (e.button) {
				case 1:													// Nur linke Maustaste ...
					switch (e.count) {
					case 1:												// ... und nur beim ersten Klick
						fileToOpen	= null;
						selectionService.setSelection(null);
						break;
					}
					break;
				}
			}

			@Override
			public void mouseDoubleClick(MouseEvent e) {
				if (fileToOpen != null) {
					fileToOpen.setValue(OpenFileActionTypes.DOUBLE_CLICK);
				}
			}
		});

		treeViewer.addSelectionChangedListener(event -> {
			FileTreeNode	node	= (FileTreeNode) ((IStructuredSelection) event.getSelection()).getFirstElement();

			// Auswahl in den SelectionService, wo er von den Popup-Handlern für das Kontextmenü erwartet wird
			selectionService.setSelection(event.getSelection());

			// Hier muss SWT mal wieder ausgetrickst werden, weil die Event-Reihenfolge einfach nur merkwürdig ist:
			// 1. Mouse Down             (1. Klick)
			// 2.    SelectionChanged als auch PostSelectionChanged
			// 3. Mouse Up               (1. Klick)
			// 4. Mouse Down             (2. Klick)
			// 5.    Mouse DoubleClick   (2. Klick)
			// 6. Mouse Up               (2. Klick)
			//
			// Der hier initial gesetzte Open-Type "KEY_SELECT" wird in nachfolgenden Mouse-Up-Events ggfs. nochmal überschrieben
			fileToOpen	= new SimpleEntry<>(node, OpenFileActionTypes.KEY_SELECT);

			Job	openFileJob	= Job.create("Open file ...", (ICoreRunnable) monitor -> {
				synchronizer.syncExec(() -> {
					if (fileToOpen == null || fileToOpen.getValue() == OpenFileActionTypes.RIGHT_CLICK)
						return;

					if (lastSelection != null && lastSelection instanceof FileTreeNodeFile && ((FileTreeNodeFile) lastSelection).isFile()) {
						try {
							closeFile(((FileTreeNodeFile) lastSelection).getPath());
							lastSelection	= null;
						} catch (IOException e) {
						}
					}

					lastSelection	= fileToOpen.getKey();

					if (treeViewer.isExpandable(fileToOpen.getKey()) && fileToOpen.getValue() == OpenFileActionTypes.DOUBLE_CLICK) {
						// Auf- oder zuklappen eines Folders je nach Status. Leere Folder sind im ELSE abzufangen
						if (treeViewer.getExpandedState(fileToOpen.getKey())) {
							treeViewer.collapseToLevel(fileToOpen.getKey(), 1);
						} else {
							treeViewer.expandToLevel(fileToOpen.getKey(), 1);
						}

					} else if (fileToOpen.getKey() instanceof FileTreeNodeFile && ((FileTreeNodeFile) fileToOpen.getKey()).isFile()) {
						FileTreeNodeFile	vfo	= (FileTreeNodeFile) fileToOpen.getKey();

						// Ob dieser Name gespeichert wird, oder nicht, entscheidet sich beim Empfänger
						eventBroker.post(FileTreeConstants.ADD_TO_QUICK_SELECT_LIST, vfo.getParent().getFullName());

						try {
							FileData	fileData	= new FileData(vfo.getPath(), vfo.isWritable());
							if(fileToOpen.getValue() == OpenFileActionTypes.DOUBLE_CLICK) {
								globalAppManager.openInNewWindow(fileData, false, SwitchPerspectiveTargetTypes.ANY);
							} else {
								ignoreSelectionEvent	= true;
								globalAppManager.openInWindow(window, fileData, fileTree.isAutoSwitchPerspective() ? SwitchPerspectiveTargetTypes.ANY : SwitchPerspectiveTargetTypes.NONE);
								ignoreSelectionEvent	= false;
							}
						} catch (IOException e) {
						}
					}
				});
			});

			openFileJob.schedule(OPEN_FILE_JOB_DELAY);
		});
	}

	private void autoSizeColumns() {
		if (fileTree.isAutoColumnWidth()) {
			// Passt die Breite aller Tabellenspalten an
			for (TreeColumn col: treeViewer.getTree().getColumns()) {
				if (!col.isDisposed()) {
					col.pack();
				}
			}
		}
	}

	private void colorizeCell(ViewerCell cell) {
		// Blau einfärben von Dateien in einem Archiv respektive grau einfärben versteckter Dateien.
		// Achtung: Vergleichbare Routine in den anderen LabelProvidern!
		if (cell.getElement() instanceof FileTreeNodeFile && ((FileTreeNodeFile) cell.getElement()).isInsideArchive()) {
			cell.setForeground(display.getSystemColor(SWT.COLOR_BLUE));
		}
		if (cell.getElement() instanceof FileTreeNodeFile && ((FileTreeNodeFile) cell.getElement()).isHidden()) {
			cell.setForeground(display.getSystemColor(SWT.COLOR_DARK_GRAY));
		}
	}

	private ControlListener createColumnControlListener(FileTreeColumn fileTreeColumn) {
		return new ControlListener() {

			@Override
			public void controlMoved(ControlEvent e) {
			}

			@Override
			public void controlResized(ControlEvent e) {
				fileTree.getColumnWidths().put(fileTreeColumn, ((TreeColumn) e.getSource()).getWidth());
			}
		};
	}

	private SelectionListener createColumnSelectionListener(FileTreeColumn fileTreeColumn, boolean ascending) {
		return new SelectionListener() {

			final	FileTreeColumn	myTreeColumn	= fileTreeColumn;

			@Override
			public void widgetSelected(SelectionEvent e) {
				widgetDefaultSelected(e);
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent e) {
				fileTree.setComparator(myTreeColumn, fileTree.getComparator().getFileTreeColumn() == myTreeColumn ? !fileTree.getComparator().isAscending() : ascending);

				BusyIndicator.showWhile(treeViewer.getControl().getDisplay(), () -> {
					// Datenmodell-Refresh: FilteredAndSorted-Liste zurücksetzen
					fileTree.getRootNode().refresh();

					// Jetzt auch den Baum neu lesen
					if (treeViewer != null && !treeViewer.getControl().getDisplay().isDisposed()) {
						treeViewer.setInput(fileTree.getRootNode());
//						treeViewer.refresh(fileTree.getRootNode(), false);
					}
				});
			}
		};
	}

	private void createContextMenu() {
		// WICHTIG! Ohne diesen hier generierten Link zwischen Tree und Popup-Menü im Model wird einfach gar nichts angezeigt!
		// https://stackoverflow.com/questions/70986667/creating-a-popup-menu-within-a-fragment-e4xmi-for-an-editor-at-extension-point-o
		menuService.registerContextMenu(treeViewer.getTree(), Constants.FILETREE_POPUP_MENU);
	}

	@Focus
	public void setFocus() {
		treeViewer.getTree().setFocus();
	}

	public void closeFile(Path file) throws IOException {
		globalAppManager.closeFile(window, false);
	}

	public void revealElement(FileTreeNode filePath, boolean showDocument) {
		if (filePath == null)
			return;

		synchronizer.syncExec(() -> {
//			treeViewer.refresh();
			treeViewer.expandToLevel(filePath, 1);
			if (showDocument) {
				treeViewer.setSelection(new StructuredSelection(filePath), true);
			} else {
				TreeItem	testItem	= (TreeItem) treeViewer.testFindItem(filePath);
				if (testItem != null) {
					treeViewer.getTree().setSelection(testItem);
				}
			}
		});
	}

	@Inject @Optional
	public void setBusyCursor(@UIEventTopic(FileTreeConstants.SET_BUSY_CURSOR) Boolean wait) {
		if (wait != null) {
			synchronizer.asyncExec(() -> {
				setBusyCursorInternal(wait);
			});
		}
	}

	private void setBusyCursorInternal(boolean wait) {
		if (!display.isDisposed() && !treeViewer.getControl().isDisposed()) {
			treeViewer.getControl().setCursor(wait ? display.getSystemCursor(SWT.CURSOR_WAIT) : null);
		}
	}

	@Inject @Optional
	public void refreshTree(@UIEventTopic(FileTreeConstants.REFRESH_FILE_TREE) FileTree fileTree) {
		if (fileTree != null) {
			updateTree();
		}
	}

	private void updateTree() {
		if (!display.isDisposed() && !treeViewer.getControl().isDisposed()) {
			synchronizer.syncExec(() -> {
				setBusyCursorInternal(true);
				treeViewer.refresh(true);
				setBusyCursorInternal(false);
			});
		}
	}

	@Inject @Optional
	public void refresh(@UIEventTopic(FileTreeConstants.RELOAD_FILE_OBJECT) FileTreeNode node) {
		if (node != null) {
			node.reload();
			updateVisibleTreeItem(node.getParent());
		} else {
			if (treeViewer.getStructuredSelection() != null) {
				FileTreeNode	selected	= (FileTreeNode) treeViewer.getStructuredSelection().getFirstElement();
				if (selected != null) {
					selected.reload();
					updateVisibleTreeItem(selected.getParent());
				} else {
					fileTree.getRootNode().reload();
					updateTree();
				}
			} else {
				fileTree.getRootNode().reload();
				updateTree();
			}
		}
	}

	private void updateVisibleTreeItem(FileTreeNode node) {
		if (node == treeViewer.getInput() || Arrays.asList(treeViewer.getVisibleExpandedElements()).stream().anyMatch(n -> n == node)) {
			if (!display.isDisposed() && !treeViewer.getControl().isDisposed()) {
				synchronizer.syncExec(() -> {
					setBusyCursorInternal(true);
					treeViewer.refresh(node, true);
					setBusyCursorInternal(false);
				});
			}
		}
	}

	// Dummerweise erfolgt die Method Injection bereits vor dem PostConstruct, wenn die fragliche
	// Variable bereits vorher gesetzt gewesen ist. In diesem Fall muss die Verarbeitung hier ab-
	// gebrochen und vom PostConstruct nach erfolgreicher Initialisierung erneut angestoßen werden.
	@Inject
	public void selectInFileTreeInjector(@Optional L2FileData fileData, Composite parent) {
		// Muss dieser Part erst noch initialisiert werden, weil das PostConstruct standardmäßig erst nach den Inject Methods kommt?
		if (treeViewer == null) {
			postConstruct(parent);
		}

		if (fileData != null && fileData.getFile() != null) {
			// showInDocument muss hier false sein, weil es sonst über die Kette openInWindow -> selectInFileTree -> revealDocument ->
			// singleClickEvent -> openInWindow zu einer doppelten Öffnung dieser Datei kommt. Mögliche Alternative wäre, dass sämtliche
			// Stellen, an denen eine Dateiauswahl im Baum möglich ist (u.a. RecentFilesHandler, QuickFolders, OpenAs, ...) in dem Sinne
			// umprogrammiert werden, dass sie selbst keinen GAM.openIn...() aufrufen, sondern nur die Datei in den Kontext setzen und
			// so gesehen selbst das singleClickEvent auslösen. Das bräuchte aber ebenfalls etwas Nachdenken, weil auf diesem Weg keine
			// Unterscheidung zwischen openInWindow und openInNewWindow möglich ist. Eine weitere Alternative wäre es, wie im Structure-
			// und ObjectsTree auf eine Hilfsvariable updateStartedHere zu setzen, die immer dann die Verarbeitung des PostSelection-
			// ChangedListener verhindert, wenn das Update automatisch respektive von außen erfolgte. Abert warum so umständlich, wenn
			// man über eine Einstellung showInDocument = false Gleiches ebenfalls erreicht?
			selectInFileTree(fileData.getFile().toAbsolutePath().toString(), false);
		}
	}

	public void selectInFileTree(String fileToSelect, boolean showDocument) {
		if (fileToSelect != null && !ignoreSelectionEvent) {
			try {
				FileTreeNode	fileObject	= fileTree.getNode(fileToSelect);
				refresh(fileObject);
				revealElement(fileObject, showDocument);
				autoSizeColumns();
			} catch (IOException e) {
			}
		}
	}

	@Inject @Optional
	public void refreshQuickSelectionControl(@UIEventTopic(FileTreeConstants.UPDATE_QUICK_SELECT_LIST) String[] quickSelectFolders) {
		if (quickSelectFolders != null) {
			quickSelectCombo.setItems(quickSelectFolders);
			quickSelectCombo.select(0);
		}
	}

}