package de.credo.lookup2.parts.structure;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StyledTextContent;
import org.eclipse.swt.custom.TextChangeListener;
import org.eclipse.swt.custom.TextChangedEvent;
import org.eclipse.swt.custom.TextChangingEvent;

/*
 * In Teilen kopiert von hier:
 * https://alvinalexander.com/java/jwarehouse/eclipse/org.eclipse.help.ui/src/org/eclipse/help/ui/internal/StyledLineWrapper.java.shtml
 *
 * und von der Default-Implementierung:
 * org.eclipse.swt.custom.DefaultContent
 */
class FixedLineWidthEncodedContentProvider implements StyledTextContent {

	private	List<TextChangeListener> textListeners = new ArrayList<>();
	private	char[]		textStore		= new char[0]; // stores the actual text
	private	String[]	lines			= new String[] {};
	private	int			charCount		= -1;
	private	int			lineWidth		= 111;

	public FixedLineWidthEncodedContentProvider() {
		setText("");
	}

	@Override
	public void addTextChangeListener(TextChangeListener l) {
		textListeners.add(l);
	}

	@Override
	public int getCharCount() {
		if (charCount != -1)
			return charCount;
		charCount = 0;
		for (String element : lines) {
			charCount += element.length();
		}
		return charCount;
	}

	@Override
	public String getLine(int i) {
		if (i >= lines.length || i < 0) {
			SWT.error(SWT.ERROR_INVALID_ARGUMENT);
		}
		return lines[i];
	}

	@Override
	public int getLineAtOffset(int offset) {
		if (offset >= getCharCount())
			return getLineCount() - 1;
		int count = 0;
		int line = -1;
		while (count <= offset) {
			count += getLine(++line).length();
		}
		return line;
	}

	@Override
	public int getLineCount() {
		return lines.length == 0 ? 1 : lines.length;
	}

	@Override
	public String getLineDelimiter() {
		return null;
	}

	@Override
	public int getOffsetAtLine(int line) {
		if (lines.length == 0)
			return 0;
		int offset = 0;
		for (int i = 0; i < line; i++) {
			offset += getLine(i).length();
		}
		return offset;
	}

	@Override
	public String getTextRange(int start, int end) {
		if (textStore == null)
			return "";
		if (textStore.length == 0)
			return "";
		return new String(Arrays.copyOfRange(textStore, start, end));
	}

	@Override
	public void removeTextChangeListener(TextChangeListener arg0) {
		textListeners.remove(arg0);
	}

	@Override
	public void replaceTextRange(int start, int replaceLength, String newText) {
		// check for invalid replace operations
		if (!isValidReplace(start, replaceLength, newText)) {
			SWT.error(SWT.ERROR_INVALID_ARGUMENT);
		}

		// inform listeners
		for (TextChangeListener l : textListeners) {
			l.textChanging(new TextChangingEvent(this));
		}

// TODO Unfertig. Der DefaultContent-Provider macht da ein ziemliches Fass auf
//		// first delete the text to be replaced
//		delete(start, replaceLength, lineCount(start, replaceLength) + 1);
//		// then insert the new text
//		insert(start, newText);

		// Listener informieren, in erster Linie das eigene StyledText-Widget.
		for (TextChangeListener l : textListeners) {
			l.textChanged(new TextChangedEvent(this));
		}
	}

	/**
	 * Determine whether or not the replace operation is valid. DefaultContent will
	 * not allow the /r/n line delimiter to be split or partially deleted.
	 * <p>
	 *
	 * @param start         start offset of text to replace
	 * @param replaceLength start offset of text to replace
	 * @param newText       start offset of text to replace
	 * @return a boolean specifying whether or not the replace operation is valid
	 */
	protected boolean isValidReplace(int start, int replaceLength, String newText) {
		if (replaceLength == 0) {
			// inserting text, see if the \r\n line delimiter is being split
			if (start == 0)
				return true;
			if (start == getCharCount())
				return true;
			char before = getTextRange(start - 1, 1).charAt(0);
			if (before == '\r') {
				char after = getTextRange(start, 1).charAt(0);
				if (after == '\n')
					return false;
			}
		} else {
			// deleting text, see if part of a \r\n line delimiter is being deleted
			char startChar = getTextRange(start, 1).charAt(0);
			if (startChar == '\n') {
				// see if char before delete position is \r
				if (start != 0) {
					char before = getTextRange(start - 1, 1).charAt(0);
					if (before == '\r')
						return false;
				}
			}
			char endChar = getTextRange(start + replaceLength - 1, 1).charAt(0);
			if (endChar == '\r') {
				// see if char after delete position is \n
				if (start + replaceLength != getCharCount()) {
					char after = getTextRange(start + replaceLength, 1).charAt(0);
					if (after == '\n')
						return false;
				}
			}
		}
		return true;
	}

	@Override
	public void setText(String text) {
		if (text == null) {
			text = "";
		}

		textStore = text.toCharArray();
		charCount = textStore.length;

		int lCount = (int) Math.ceil((double) charCount / lineWidth);
		lines = new String[Math.max(lCount, 1)];
		Arrays.fill(lines, "");

		char[] line = new char[Math.min(lineWidth, charCount)];
		int lc = 0;

		// Ersetzen aller bekannten Control Character durch eine passende Ersatzdarstellung
		// und direkt anschließendes Aufteilen in Zeilen
		for (int c = 0; c < charCount; c++) {
			// Leider gibt es einen Haufen Zeichen, die zwar korrekt konvertiert werden können,
			// aber leider trotzdem nicht zur Anzeige in einem Hex-Modus taugen. Die werden hier
			// noch einmal von Hand umgeschossen.
			// Dabei handelt es sich z.B. um die Control Characters, die ziemlich vollständig
			// hier gelistet sind: https://www.aivosto.com/articles/control-characters.html
			// (Titel: "Control characters in ASCII and Unicode")
			if (textStore[c] < 0x20								// C0 Control Characters 0x00 - 0x1F
			 || textStore[c] > 0x7E && textStore[c] < 0xA0		// C1 Control Characters 0x80 - 0x9F + DEL (0x7F)
			 || textStore[c] == 0x2028							// Unicode LSEP Line Separator
			 || textStore[c] == 0x2029) {						// Unicode PSEP Paragraph Separator
				textStore[c] = '.';

			} else if (textStore[c] == 0xA0) {					// 8859 NBSP No Break Space
				textStore[c] = ' ';

			} else if (textStore[c] == 0xAD) {					// 8859 SHY Soft-Hyphen
				textStore[c] = '-';
			}

			int l = c / lineWidth;
			if (l > lc) {
				lines[lc] = new String(line);
				line = new char[Math.min(lineWidth, charCount - c)];
				lc = l;
			}

			line[c % lineWidth] = textStore[c];
		}

		lines[lc] = new String(line);

		// Listener informieren, in erster Linie das eigene StyledText-Widget. Ohne
		// Listener wird sonst alles bei einer Textänderung nicht richtig initialisiert.
		for (TextChangeListener l : textListeners) {
			l.textSet(new TextChangedEvent(this));
		}
	}

	public int getLineWidth() {
		return lineWidth;
	}

	public void setLineWidth(int lineWidth) {
		this.lineWidth = lineWidth;
	}

}
