package de.credo.lookup2.parts.structure;

import java.nio.ByteBuffer;
import java.nio.charset.CharacterCodingException;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CodingErrorAction;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.LinkedList;
import java.util.List;

import jakarta.annotation.PostConstruct;
import jakarta.inject.Inject;

import org.eclipse.e4.core.contexts.IEclipseContext;
import org.eclipse.e4.core.di.annotations.Optional;
import org.eclipse.e4.core.services.events.IEventBroker;
import org.eclipse.e4.ui.di.UISynchronize;
import org.eclipse.e4.ui.model.application.ui.basic.MPart;
import org.eclipse.e4.ui.model.application.ui.basic.MTrimmedWindow;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.custom.StyleRange;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.dnd.DND;
import org.eclipse.swt.dnd.DropTarget;
import org.eclipse.swt.dnd.DropTargetEvent;
import org.eclipse.swt.dnd.DropTargetListener;
import org.eclipse.swt.dnd.FileTransfer;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.ControlListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;

import de.credo.common.bundles.messages.CdoLogger;
import de.credo.lookup2.Constants;
import de.credo.lookup2.GlobalAppManager;
import de.credo.lookup2.SwitchPerspectiveTargetTypes;
import de.credo.lookup2.parsers.api.L2ParserConfig;
import de.credo.lookup2.parsers.api.SingleByteCharsets;
import de.credo.lookup2.parsers.api.streams.L2DataStream;
import de.credo.lookup2.parsers.api.structure.L2NoHex;
import de.credo.lookup2.parsers.api.structure.L2ParentElement;
import de.credo.lookup2.parsers.api.structure.L2RichContainer;
import de.credo.lookup2.parsers.api.structure.L2RichParentElement;
import de.credo.lookup2.plugins.api.L2Exception;
import de.credo.lookup2.plugins.impl.FileData;

public class HexViewPart {

	private	final static	CdoLogger	LOGGER	= new CdoLogger(Constants.LOGGER_ORIGIN_HEXVIEW);

	private final static	char[]	HEX_CHARS	= "0123456789ABCDEF".toCharArray();
	private	final static	int		COL_GAP		= 10;
	private	final static	int		MARGIN		= 5;
	private	final static	int		MAX_LENGTH	= 16384;	// 16 kib + x Overflow
	private	final static	Color[]	DEFAULT_COLORS	= new Color[] {
			                                          Display.getDefault().getSystemColor(SWT.COLOR_BLUE),
			                                          Display.getDefault().getSystemColor(SWT.COLOR_RED),
			                                          Display.getDefault().getSystemColor(SWT.COLOR_DARK_GREEN),
			                                          Display.getDefault().getSystemColor(SWT.COLOR_DARK_YELLOW)};
	private	final static	Color	DARK_GRAY		= Display.getDefault().getSystemColor(SWT.COLOR_DARK_GRAY);

	@Inject	IEclipseContext		context;
	@Inject	UISynchronize		synchronizer;
	@Inject	Display				display;
	@Inject	IEventBroker		eventBroker;
	@Inject	MTrimmedWindow		window;
	@Inject	GlobalAppManager	globalAppManager;
	@Inject	MPart				hexPart;
	@Optional
	@Inject	L2ParentElement		parentElement;
	@Optional
	@Inject	Charset				encoding;

	private	Composite			parent;
	private	ScrolledComposite	scrollable;
	private	Composite			composite;

	private	StyledText			addrText;
	private	GridData			addrLayout	= new GridData(SWT.LEFT, SWT.FILL, true, false);
	private	StyledText			hexText;
	private	GridData			hexLayout	= new GridData(SWT.LEFT, SWT.FILL, true, false);
	private	StyledText			encText;
	private	GridData			encLayout	= new GridData(SWT.LEFT, SWT.FILL, true, false);
	private	int					charWidth;
	private	FixedLineWidthHexValuesContentProvider	fixedLineWidthHexValuesContent	= new FixedLineWidthHexValuesContentProvider();
	private	FixedLineWidthEncodedContentProvider	fixedLineWidthEncodedContent	= new FixedLineWidthEncodedContentProvider();
	private	List<ExtStyleRange>	ranges;


	private	int					lineWidth	= 20;
	private	CharsetDecoder		decoder;
	private	long				startByte;
	private	byte[]				byteArray;

	@PostConstruct
	public void postConstruct(Composite parent) {
		this.parent	= parent;

		createControls();

		// Händisches Nachholen der möglicherweise zu früh erfolgten Method Injection unten
		if (parentElement != null) {
			showHex(parentElement);
		}
	}

	private void createControls() {
		scrollable	= new ScrolledComposite(parent, SWT.V_SCROLL);
		scrollable.setAlwaysShowScrollBars(true);
		scrollable.setExpandHorizontal(true);
		scrollable.setExpandVertical(true);
		scrollable.setMinWidth(0);
		scrollable.setMinHeight(0);

		scrollable.addControlListener(new ControlListener() {

			@Override
			public void controlResized(ControlEvent e) {
				if (!composite.isDisposed()) {
					int	cpWidth	= ((Composite) e.widget).getClientArea().width;	// Nutzbare Breite des gesamten Composite
					cpWidth		= cpWidth - 8 * charWidth;						// ... abzüglich fester Breite der ersten Spalte
					cpWidth		= cpWidth - 2 * COL_GAP;						// ... abzüglich zweier Mindestabstände zwischen den drei Spalten
					cpWidth		= cpWidth - 6 * MARGIN;							// ... abzüglich 6 (möglicherweise unsichtbaren) Linien, 2 je Spalte
					int	lWidth	= cpWidth / (4 * charWidth);					// geteilt durch jeweils vier Zeichen pro anzuzeigendem Byte ("5A " [hex] + "X" [enc]) und abgerundet

					if (lWidth != lineWidth) {
						lineWidth	= lWidth;

						addrLayout.minimumWidth	= 8 * charWidth;
						hexLayout.minimumWidth	= lineWidth * 3 * charWidth;
						encLayout.minimumWidth	= lineWidth * 1 * charWidth;

						composite.layout();

						// Reformatieren auch des Inhalts bei einer Größenänderung
						if (byteArray != null) {
							if (encText.getRanges().length > 1) {
								presentBytes(lineWidth, Math.max(0, encText.getRanges()[encText.getRanges().length - 2] - 2 * lineWidth));
							} else {
								presentBytes(lineWidth, 0);
							}
						}
					}
				}
			}

			@Override
			public void controlMoved(ControlEvent e) {
			}
		});

		composite	= new Composite(scrollable, SWT.NONE);
		GridLayout	gl			= new GridLayout(3, false);
		gl.horizontalSpacing	= COL_GAP;
		gl.marginWidth			= MARGIN;
		composite.setLayout(gl);
//		composite.setBackground(display.getSystemColor(SWT.COLOR_WHITE));

// SWT.WRAP entfernt, weil WordWrap bei StyledText extreme Performanceeinbußen bringt. Da gibt
// es massig Bugzilla-Einträge zu. Scheint in unserem Fall aber auch gar nicht nötig zu sein?!
//		addrText	= new StyledText(composite, SWT.LEFT | SWT.MULTI | SWT.READ_ONLY | SWT.WRAP);
		addrText	= new StyledText(composite, SWT.LEFT | SWT.MULTI | SWT.READ_ONLY);
		addrText.setLayoutData(addrLayout);
//		addrText.setBackground(display.getSystemColor(SWT.COLOR_WIDGET_BACKGROUND));
		addrText.setFont(JFaceResources.getFontRegistry().getBold(JFaceResources.TEXT_FONT));

// SWT.WRAP entfernt, weil WordWrap bei StyledText extreme Performanceeinbußen bringt. Da gibt
// es massig Bugzilla-Einträge zu. Scheint in unserem Fall aber auch gar nicht nötig zu sein?!
//		hexText		= new StyledText(composite, SWT.LEFT | SWT.MULTI | SWT.READ_ONLY | SWT.WRAP);
		hexText		= new StyledText(composite, SWT.LEFT | SWT.MULTI | SWT.READ_ONLY);
		hexText.setLayoutData(hexLayout);
//		hexText.setBackground(display.getSystemColor(SWT.COLOR_WIDGET_BACKGROUND));
		hexText.setFont(JFaceResources.getTextFont());
		hexText.setContent(fixedLineWidthHexValuesContent);

		hexText.addSelectionListener(new SelectionListener() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				Point	selection	= hexText.getSelection();
				int		startByte	= selection.x / 3;
				int		endByte		= (selection.y + 1) / 3;
				encText.setSelection(new Point(startByte, endByte));
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent e) {
				widgetSelected(e);
			}
		});

// SWT.WRAP entfernt, weil WordWrap bei StyledText extreme Performanceeinbußen bringt. Da gibt
// es massig Bugzilla-Einträge zu. Scheint in unserem Fall aber auch gar nicht nötig zu sein?!
//		encText		= new StyledText(composite, SWT.LEFT | SWT.MULTI | SWT.READ_ONLY | SWT.WRAP);
		encText		= new StyledText(composite, SWT.LEFT | SWT.MULTI | SWT.READ_ONLY);
		encText.setLayoutData(encLayout);
//		encText.setBackground(display.getSystemColor(SWT.COLOR_WIDGET_BACKGROUND));
		encText.setFont(JFaceResources.getFontRegistry().getBold(JFaceResources.TEXT_FONT));
		encText.setContent(fixedLineWidthEncodedContent);

		encText.addSelectionListener(new SelectionListener() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				Point	selection	= encText.getSelection();
				int		startByte	= selection.x * 3;
				int		endByte		= selection.y * 3 - (selection.y > selection.x ? 1 : 0);
				hexText.setSelection(new Point(startByte, endByte));
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent e) {
				widgetSelected(e);
			}
		});

		// Berechnen der Zeichenbreite im gewählten Font zur späteren Berechnung der möglichen Bytes pro Zeile
		GC		gc	= new GC(addrText);
		Point	cw	= gc.textExtent("0QWM8");
		charWidth	= (int) Math.ceil((double) cw.x / 5);
		gc.dispose();

		scrollable.setContent(composite);

		DropTarget	target	= new DropTarget(composite, DND.DROP_COPY | DND.DROP_LINK | DND.DROP_MOVE);
		target.setTransfer(new Transfer[] {FileTransfer.getInstance()});
		target.addDropListener(new DropTargetListener() {

			@Override	public void dropAccept(DropTargetEvent event) {}

			@Override
			public void drop(DropTargetEvent event) {
				if (event.data instanceof String[]) {
					String[]	filenames	= (String[]) event.data;
					if (filenames.length > 0) {
						for (int i = 0; i < filenames.length; i++) {
							java.nio.file.Path	filePath	= Paths.get(filenames[i]);
							globalAppManager.openInNewWindow(new FileData(filePath, Files.isWritable(filePath)), true, SwitchPerspectiveTargetTypes.ANY);
						}
					}
				}
			}

			@Override	public void dragOver(DropTargetEvent event) {}

			@Override	public void dragOperationChanged(DropTargetEvent event) {}

			@Override	public void dragLeave(DropTargetEvent event) {}

			@Override	public void dragEnter(DropTargetEvent event) {}
		});
	}

	@Inject
	public void showHex(@Optional final L2ParentElement parentElement) {
		// Wurde dieser Part wegen "falscher" Method-Injection-Reihenfolge noch nicht initialisiert, dann wird
		// hier abgebrochen und die Übergabe lediglich zwischengespeichert. Die Routine wird dann später (nach
		// der Initialisierung) aus dem PostConstruct noch einmal neu aufgerufen.
		if (this.parent == null)
			return;

		if (parentElement == null)
			return;

		synchronizer.asyncExec(() -> {
			L2ParentElement workElement	= parentElement;

			if (workElement instanceof L2NoHex || workElement.getChunk() == null) {
				composite.setVisible(false);
				return;
			} else {
				composite.setVisible(true);
			}

			// Ein- bzw. erstmaliges Setzen des Default-Encodings beim ersten Aufruf
			if (encoding == null) {
				Charset	newEncodingFromParser	= workElement.getContextVariable(Charset.class);
				if (newEncodingFromParser == null) {
					newEncodingFromParser	= StandardCharsets.ISO_8859_1;
				}
				context.set(Charset.class, newEncodingFromParser);
			}

			L2ParserConfig	config		= workElement.getContextVariable(L2ParserConfig.class);
			Color[]			colors		= config != null ? config.getHexWindowColors() != null ? config.getHexWindowColors() : DEFAULT_COLORS : DEFAULT_COLORS;
			L2DataStream	stream		= workElement.getChunk().getStream();
			startByte	= stream.getStartByte();
			ranges	= new LinkedList<>();

			// Sollte der Stream geschlossen sein, passiert hier einfach nichts
			if (stream.isOpen()) {
				// Container werden einfach im Ganzen, ohne Sonderbehandlung (bis auf mögliches Kürzen) angezeigt,
				// gleiches gilt für Records, die den NoHexMarker gesetzt haben
				if (workElement instanceof L2RichContainer) {
					try {
						long	containerStartByte	= workElement.getChunk().getStartByte();
						long	containerSize		= workElement.getChunk().getExtend();
						startByte	= Math.max(startByte, containerStartByte - 8 * lineWidth);				// 0 = kleinster möglicher Startwert im Stream
						int	length	= (int) Math.ceil((double) (containerStartByte - startByte + containerSize + 8 * lineWidth) / lineWidth) * lineWidth;
						length		= Math.min(length, MAX_LENGTH);											// Einkürzen auf maxLength, wenn nötig
						length		= (int) (Math.min(stream.getSize(), startByte + length) - startByte);	// Einkürzen auf die nur noch bis Dateiende zur Verfügung stehenden Bytes, wenn nötig

						ranges.add(0,
								new ExtStyleRange((int) (containerStartByte - startByte),
										(int) Math.min(containerSize, length + startByte - containerStartByte),
										colors[0],
										null,
										containerStartByte));

						byteArray	= stream.getBytes(startByte, length);

						presentBytes(lineWidth, Math.max(0, ranges.get(0).start - 2 * lineWidth));
					} catch (L2Exception e) {
						LOGGER.fatal(e.getLocalizedMessage(), e);
					}

					// Records kriegen eine komplizierte geschachtelte Struktur an Farben zur Unterscheidung. Das gilt
					// alles jedoch nur für Records, die einen für den Benutzer sinnvollen und keinen rein technischen
					// Hintergrund haben, mithin den NoHexMarker eben *nicht* gesetzt haben
				} else {
					try {
						// Zusammentragen der verschiedenen Hierarchien an Records, von innen nach außen.
						int	cIndex	= 0;
						while (workElement != null) {
							if (!(workElement instanceof L2RichContainer) && workElement.isVisible() && workElement.getChunk().getStream().equals(stream)) {
								// Erst den normalen Datenteil des Records ...
								ranges.add(0,
										new ExtStyleRange(0,
												(int) workElement.getChunk().getExtend(),
												colors[cIndex],
												null,
												workElement.getChunk().getStartByte()));
								cIndex++;
								if (cIndex == colors.length) {
									cIndex	= 0;
								}

								// ... und, wenn es gibt, noch Vorlauf und Nachlauf separat markiert
								if (workElement instanceof L2RichParentElement<?>) {
									L2RichParentElement<?>	richFolder	= (L2RichParentElement<?>) workElement;

									if (richFolder.getLength().dataOffset() + richFolder.getLength().tailLength() > 0) {
										ranges.add(0,
												new ExtStyleRange(0,
														(int) (richFolder.getChunk().getExtend() + richFolder.getLength().tailLength()),
														DARK_GRAY,
														null,
														richFolder.getChunk().getStartByte() - richFolder.getLength().dataOffset()));
									}
								}
							}
							workElement	= workElement.getParent();
						}

						// Neuberechnen aller Startwerte und Längen der verschiedenen ineinander geschachtelten StyleRanges.
						// Zusätzlich wird ein neutraler äußerer Rahmen von 8 Zeilen hinzugefügt, ohne eigenen StyleRange,
						// nur über Offsets.
						// Bei der Gelegenheit werden dann auch gleich möglicherweise ungültige Werte korrigiert

						// 8 Zeilen Vorlauf und Sicherstellen, dass man damit nicht in Negative läuft
						startByte	= Math.max(startByte, ranges.get(0).startByte - 8 * lineWidth);
						// Sicherstellen, dass das gesuchte Element im Anzeigebereich ist (zur Not wird vorne weiter abgeschnitten)
						startByte	= Math.max(startByte, ranges.get(ranges.size() - 1).startByte - MAX_LENGTH / 3 * 2);
						// Justieren auf ganze Zeilen
						int	length	= (int) Math.ceil((double) (ranges.get(0).startByte - startByte + ranges.get(0).length + 8 * lineWidth) / lineWidth) * lineWidth;
						// Einkürzen auf maxLength, wenn nötig
						length		= Math.min(length, MAX_LENGTH);
						// Einkürzen auf die nur noch bis Dateiende zur Verfügung stehenden Bytes, wenn nötig
						length		= (int) (Math.min(stream.getSize(), startByte + length) - startByte);
						for (ExtStyleRange esr: ranges) {
							esr.start	= (int) (esr.startByte - startByte);
							esr.length	= Math.min(esr.length, length - esr.start);
						}
						byteArray	= stream.getBytes(startByte, length);

						presentBytes(lineWidth, Math.max(0, ranges.get(ranges.size() - 1).start - 2 * lineWidth));

					} catch (L2Exception e) {
						LOGGER.fatal(e.getLocalizedMessage(), e);
					}
				}
			} else {
				// Leeren des Fensters
				addrText.setText("");
				hexText.setText("");
				encText.setText("");
			}
		});
	}

	private void presentBytes(int lineWidth, int cursorPosition) {
		presentBytesAddressColumn(lineWidth);
		presentBytesHexValuesColumn(lineWidth);
		presentBytesEncodedColumn(lineWidth);

		// Höhenänderung des Scrollbalkens
		composite.layout();
		scrollable.setMinSize(composite.computeSize(scrollable.getClientArea().width, SWT.DEFAULT));

		// Cursorposition setzen und ScrolledComposite passend scrollen
		try {
			hexText.setSelection(cursorPosition);
			encText.setSelection(cursorPosition);
			scrollable.setOrigin(encText.getLocationAtOffset(cursorPosition));
		} catch (IllegalArgumentException e) {		// Index out of bounds, kann vorkommen wegen des abgeschnittenen Textes
			// Tue nichts
		}
	}

	private void presentBytesAddressColumn(int lineWidth) {
		StringBuilder	address	= new StringBuilder();
		for (long addr = 0; addr < byteArray.length; addr += lineWidth) {
			address.append(String.format("%08x", startByte + addr)).append(System.lineSeparator());
		}
		if (!addrText.isDisposed()) {
			addrText.setText(address.toString());
		}
	}

	private void presentBytesHexValuesColumn(int lineWidth) {
		// Der Text wird vom oben mitgegebenen StyleTextContent zur Anzeige noch mal modifiziert. Der
		// hier bricht zum Beispiel die Anzeige hart nach lineWidth * 3 Zeichen um.
		fixedLineWidthHexValuesContent.setLineWidth(lineWidth);
		if (!hexText.isDisposed()) {
			hexText.setText(bytesToHexString(byteArray) + (byteArray.length == MAX_LENGTH ? ">>>" : ""));

			// Dynamisches Erweitern der StyleRanges. Jeder Positionswert muss für die Hexanzeige mit drei multipliziert werden
			for (int i = 0; i < ranges.size(); i++) {
				ExtStyleRange	esr	= ranges.get(i);
				if (esr.start < MAX_LENGTH) {
					try {
						hexText.setStyleRange(new StyleRange(esr.start * 3, esr.length * 3, esr.foreground, esr.background, esr.fontStyle));
					} catch (IllegalArgumentException e) {
						// Tue nichts. Der Fehler kann auftreten, wenn die zu markierenden Bereiche nicht
						// sauber ineinander liegen, zum Beispiel wenn ein Längenbyte 0 gewesen ist.
					}
				}
			}
		}
	}

	private void presentBytesEncodedColumn(int lineWidth) {
		try {
			// Der Text wird vom oben mitgegebenen StyleTextContent zur Anzeige noch mal modifiziert. Unter
			// anderem werden die nicht darstellbaren Control Character ersetzt, viel wichtiger ist aber
			// die alternative Zeilenumbruchmethode an fester Zeilenlänge. Der "DefaultContent"-Provider
			// bricht nämlich immer zuerst an Leerzeichen um (SWT.WRAP), was zu Verschiebungen führt, wenn
			// im darzustellenden Text gewollte Leerzeichen enthalten sind.
			fixedLineWidthEncodedContent.setLineWidth(lineWidth);
			if (!encText.isDisposed()) {
				encText.setText(decoder.decode(ByteBuffer.wrap(byteArray)).toString());
				for (ExtStyleRange esr: ranges) {
					if (esr.start < MAX_LENGTH) {
						try {
							encText.setStyleRange(esr);
						} catch (IllegalArgumentException e) {
							// Tue nichts. Der Fehler kann auftreten, wenn die zu markierenden Bereiche nicht
							// sauber ineinander liegen, zum Beispiel wenn ein Längenbyte 0 gewesen ist.
						}
					}
				}
			}

		} catch (CharacterCodingException e) {
			// Im Fehlerfall Löschen der Anzeige
			encText.setText("----");
			LOGGER.error(e.getLocalizedMessage(), e);
		}
	}

	public static String bytesToHexString(byte[] bytes) {
		if (bytes == null || bytes.length == 0)
			return "";

		char[]	hexChars	= new char[bytes.length * 3];

		for (int i = 0; i < bytes.length; i++) {
			int	j	= i * 3;
			int v	= bytes[i] & 0xFF;
			hexChars[j]		= HEX_CHARS[v >>> 4];
			hexChars[j + 1]	= HEX_CHARS[v & 0x0F];
			hexChars[j + 2]	= ' ';
		}

		return new String(hexChars);
	}

	@Inject
	public void changeEncoding(@Optional Charset newEncoding) {
		// Wurde dieser Part wegen "falscher" Method-Injection-Reihenfolge noch nicht initialisiert, dann wird
		// hier abgebrochen und die Übergabe lediglich zwischengespeichert. Die Routine wird dann später (nach
		// der Initialisierung) aus dem PostConstruct noch einmal neu aufgerufen.
		if (this.parent == null)
			return;

		if (newEncoding == null)
			return;

		// Testen, ob das übergebene Encoding in der Liste der unterstützten Encodings ist, sonst umschießen
		encoding	= SingleByteCharsets.from(newEncoding) != null ? newEncoding : StandardCharsets.ISO_8859_1;

		decoder	= encoding.newDecoder();
		decoder.onMalformedInput(CodingErrorAction.REPLACE);
		decoder.onUnmappableCharacter(CodingErrorAction.REPLACE);
		decoder.replaceWith(".");

		// Reformatieren des Inhalts bei neuem Encoding
		if (byteArray != null) {
			presentBytesEncodedColumn(lineWidth);
		}
	}

}