package de.credo.lookup2.prefpages;

import java.io.IOException;
import java.net.URL;
import java.util.Locale;

import jakarta.inject.Inject;

import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.e4.core.contexts.ContextInjectionFactory;
import org.eclipse.e4.core.contexts.IEclipseContext;
import org.eclipse.e4.core.di.annotations.Optional;
import org.eclipse.e4.core.services.nls.Translation;
import org.eclipse.e4.ui.workbench.IWorkbench;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.preference.PreferencePage;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.swt.SWT;
import org.eclipse.swt.dnd.Clipboard;
import org.eclipse.swt.dnd.TextTransfer;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Spinner;
import org.eclipse.swt.widgets.Text;
import org.osgi.framework.Bundle;
import org.osgi.framework.FrameworkUtil;
import org.osgi.service.component.annotations.Component;

import de.credo.common.bundles.EclipseINI;
import de.credo.common.bundles.preferencepages.MyPrefPage;
import de.credo.lookup2.GlobalAppManager;
import de.credo.lookup2.Messages;
import de.credo.lookup2.plugins.impl.PerspectiveTypes;

@Component
public class StartupSettingsPrefPage extends PreferencePage implements MyPrefPage {

	private	final static	IPath	TOOLTIP_IMAGE_PATH	= new Path("src/main/resources/icons/_16/question.png");
	private	final static	String	TOOLTIP_IMAGE_KEY	= "TOOLTIP_IMAGE_KEY";

	public enum AvailableLanguages {
		AUTO	(null),
		ENGLISH	(Locale.ENGLISH),
		DEUTSCH	(Locale.GERMAN),
		;

		private	Locale	locale;

		AvailableLanguages(Locale locale) {
			this.locale	= locale;
		}

		public Locale getLocale() {
			return locale;
		}

		public String getDisplayLanguage() {
			return locale != null ? locale.getDisplayLanguage(locale) : null;
		}

		public static AvailableLanguages fromLocale(Locale language) {
			for (AvailableLanguages al: AvailableLanguages.values()) {
				if (al.getLocale() != null
						&& al.getLocale().equals(language))
					return al;
			}
			return AvailableLanguages.AUTO;
		}

		public static AvailableLanguages fromDisplayLanguage(String displayLanguage) {
			for (AvailableLanguages al: AvailableLanguages.values()) {
				if (al.getLocale() != null
						&& al.getLocale().getDisplayLanguage(al.getLocale()).equals(displayLanguage))
					return al;
			}
			return AvailableLanguages.AUTO;
		}
	}

	public enum StartPerspectives {
		EMPTY	(PerspectiveTypes.EMPTY),
		FILES	(PerspectiveTypes.FILES),
		;

		private	PerspectiveTypes	perspectiveType;

		StartPerspectives(PerspectiveTypes perspectiveType) {
			this.perspectiveType	= perspectiveType;
		}

		public PerspectiveTypes getPerspectiveType() {
			return perspectiveType;
		}

		public static StartPerspectives from(String perspectiveName) {
			for (StartPerspectives sp: StartPerspectives.values()) {
				if (Messages.getString(sp.getPerspectiveType().getMessageKey()).equals(perspectiveName))
					return sp;
			}
			return StartPerspectives.FILES;
		}

		public static StartPerspectives from(PerspectiveTypes perspectiveType) {
			for (StartPerspectives sp: StartPerspectives.values()) {
				if (sp.getPerspectiveType() == perspectiveType)
					return sp;
			}
			return StartPerspectives.FILES;
		}
	}

	@Inject	EclipseINI			eclipseINI;
	@Inject	Display				display;
	@Inject	IEclipseContext		context;
	@Optional
	@Inject	IWorkbench			workbench;
	@Optional
	@Inject	GlobalAppManager	globalAppManager;
	@Inject	@Translation	Messages	messages;

	private	Composite		settingsComposite;
	private	Combo			appLanguageCombo;
	private	int				appLanguageComboSelection	= -1;
	private	Spinner			maxHeapSpinner;
	private	Combo			twoStepCloseCombo;
	private	Combo			startPerspectiveCombo;
	private	Label			restartHintLabel;

	@Override
	public String getTitle() {
		return messages.startupSettingsPreferencePageTitle;
	}

	@Override
	public int orderKey() {
		return 110;
	}

	@Override
	public String[] path() {
		return new String[] {messages.startupSettingsPreferencePageTitle};
	}

	@Override
	public boolean creatable() {
		return true;
	}

	@Override
	protected Control createContents(Composite parent) {
// Test-Code, da in Eclipse IDE Debug keine INI-Datei existiert!
//eclipseINI.setIniFile(Paths.get("C:\\Projekte\\Lookup2\\de.credo.lookup2.product\\target\\products\\Lookup2\\win32\\win32\\x86_64\\Lookup2\\Lookup2.ini"));
//eclipseINI.load();

		Composite	composite	= new Composite(parent, SWT.NONE);
		GridLayout	pageLayout	= new GridLayout(1, true);
		pageLayout.verticalSpacing	= 10;
		pageLayout.marginBottom		= 100;
		composite.setLayout(pageLayout);

		settingsComposite	= new Composite(composite, SWT.NONE);
		settingsComposite.setLayoutData(new GridData(SWT.FILL, SWT.BEGINNING, true, true));
		GridLayout	settingsLayout		= new GridLayout(2, false);
		settingsLayout.marginTop		= 10;
		settingsLayout.verticalSpacing	= 15;
		settingsComposite.setLayout(settingsLayout);

		createSettings(settingsComposite);
		createHintComposite(composite);

		performDefaults();
		updateGUI();

		return composite;
	}

	protected void createSettings(Composite settingsComposite) {
		Label	appLanguageLabel	= new Label(settingsComposite, SWT.NONE);
		appLanguageLabel.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));
		appLanguageLabel.setText(messages.startupSettingsPreferencePageAppLanguage);

		appLanguageCombo	= new Combo(settingsComposite, SWT.DROP_DOWN | SWT.READ_ONLY);
		appLanguageCombo.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		String[]	languageSelectionLabels	= new String[AvailableLanguages.values().length];
		for (int l = 0; l < AvailableLanguages.values().length; l++) {
			String	displayLanguage	= AvailableLanguages.values()[l].getDisplayLanguage();
			if (displayLanguage == null) {
				displayLanguage	= messages.startupSettingsPreferencePageAppLanguageAutomatic;
			}
			languageSelectionLabels[l]	= displayLanguage;
		}
		appLanguageCombo.setItems(languageSelectionLabels);

		appLanguageCombo.addSelectionListener(new SelectionListener() {

			@Override
			public void widgetSelected(SelectionEvent e) {
				appLanguageComboSelection	= appLanguageCombo.getSelectionIndex();
				updateGUI();
			}

			@Override
			public void widgetDefaultSelected(SelectionEvent e) {
				widgetSelected(e);
			}
		});

		Composite	maxHeapLabelComposite	= new Composite(settingsComposite, SWT.NONE);
		maxHeapLabelComposite.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, false, false));
		GridLayout	mhlcgl	= new GridLayout(2, false);
		mhlcgl.marginWidth	= 0;
		mhlcgl.marginHeight	= 0;
		maxHeapLabelComposite.setLayout(mhlcgl);

		Label	maxHeapLabel	= new Label(maxHeapLabelComposite, SWT.NONE);
		maxHeapLabel.setLayoutData(new GridData(SWT.BEGINNING, SWT.BOTTOM, false, false));
		maxHeapLabel.setText(messages.startupSettingsPreferencePageMaxHeapLabel);

		Image	icon	= JFaceResources.getImageRegistry().get(TOOLTIP_IMAGE_KEY);
		if (icon == null) {
			Bundle	bundle	= FrameworkUtil.getBundle(getClass());
			URL		imgFile	= FileLocator.find(bundle, TOOLTIP_IMAGE_PATH, null);
			icon	= ImageDescriptor.createFromURL(imgFile).createImage();
			JFaceResources.getImageRegistry().put(TOOLTIP_IMAGE_KEY, icon);
		}
		Label	maxHeapTooltipLabel	= new Label(maxHeapLabelComposite, SWT.NONE);
		maxHeapTooltipLabel.setLayoutData(new GridData(SWT.BEGINNING, SWT.TOP, false, false));
		maxHeapTooltipLabel.setImage(icon);
		maxHeapTooltipLabel.setToolTipText(messages.startupSettingsPreferencePageMaxHeapTooltip);

		Composite	maxHeapValueComposite	= new Composite(settingsComposite, SWT.NONE);
		maxHeapValueComposite.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		GridLayout	mhvcgl	= new GridLayout(2, false);
		mhvcgl.marginWidth	= 0;
		mhvcgl.marginHeight	= 0;
		maxHeapValueComposite.setLayout(mhvcgl);

		maxHeapSpinner	= new Spinner(maxHeapValueComposite, SWT.RIGHT);
		GridData	mhgd	= new GridData(SWT.LEFT, SWT.CENTER, false, false);
		mhgd.widthHint		= 20;
		maxHeapSpinner.setLayoutData(mhgd);
		maxHeapSpinner.setDigits(1);
		maxHeapSpinner.setIncrement((int) (0.5 * 10));
		maxHeapSpinner.setMinimum( 4 * 10);
		maxHeapSpinner.setMaximum(64 * 10);
		maxHeapSpinner.addModifyListener(e -> {
			updateGUI();
		});

		Label	maxHeapUnitLabel	= new Label(maxHeapValueComposite, SWT.NONE);
		maxHeapUnitLabel.setLayoutData(new GridData(SWT.LEFT, SWT.CENTER, false, false));
		maxHeapUnitLabel.setText(messages.startupSettingsPreferencePageMaxHeapUnitLabel);

		Label	twoStepCloseLabel	= new Label(settingsComposite, SWT.NONE);
		twoStepCloseLabel.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));
		twoStepCloseLabel.setText(messages.startupSettingsPreferencePageTwoStepCloseLabel);

		twoStepCloseCombo	= new Combo(settingsComposite, SWT.DROP_DOWN | SWT.READ_ONLY);
		twoStepCloseCombo.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		twoStepCloseCombo.setItems(messages.startupSettingsPreferencePageTwoStepCloseDelayed,
		                           messages.startupSettingsPreferencePageTwoStepCloseImmediately);

		Label	startPerspectiveLabel	= new Label(settingsComposite, SWT.NONE);
		startPerspectiveLabel.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, false, false));
		startPerspectiveLabel.setText(messages.startupSettingsPreferencePageStartPerspectiveLabel);

		startPerspectiveCombo	= new Combo(settingsComposite, SWT.DROP_DOWN | SWT.READ_ONLY);
		startPerspectiveCombo.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
		startPerspectiveCombo.setItems(Messages.getString(StartPerspectives.values()[0].getPerspectiveType().getMessageKey()),
				                       Messages.getString(StartPerspectives.values()[1].getPerspectiveType().getMessageKey()));
	}

	protected void createHintComposite(Composite generalComposite) {
		restartHintLabel	= new Label(generalComposite, SWT.NONE);
		restartHintLabel.setFont(JFaceResources.getFontRegistry().getDescriptor(JFaceResources.BANNER_FONT).increaseHeight(+1).withStyle(SWT.BOLD).createFont(generalComposite.getDisplay()));
		restartHintLabel.setForeground(Display.getDefault().getSystemColor(SWT.COLOR_RED));
		restartHintLabel.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, false));
		restartHintLabel.setText(Messages.getString("STARTUP_SETTINGS_RESTART_HINT"));
	}

	protected void updateGUI() {
		if (appLanguageComboSelection > -1 && appLanguageComboSelection != AvailableLanguages.fromLocale(eclipseINI.getLanguage()).ordinal()
		 || 1024 * maxHeapSpinner.getSelection() / 10 != eclipseINI.getMaxHeap()) {
			if (appLanguageComboSelection > -1 && appLanguageComboSelection != AvailableLanguages.fromLocale(eclipseINI.getLanguage()).ordinal()) {
				Locale	selectedLanguage	= AvailableLanguages.values()[appLanguageComboSelection].getLocale();
				restartHintLabel.setText(Messages.getString("STARTUP_SETTINGS_RESTART_HINT", selectedLanguage));
			}
			restartHintLabel.setVisible(true);
		} else {
			restartHintLabel.setVisible(false);
		}
		restartHintLabel.requestLayout();
	}

	@Override
	protected void performDefaults() {
		appLanguageCombo.select(AvailableLanguages.fromLocale(eclipseINI.getLanguage()).ordinal());
		maxHeapSpinner.setSelection((int)((double) eclipseINI.getMaxHeap() / 1024 * 2 + 0.5) / 2 * 10);
		twoStepCloseCombo.select(globalAppManager.getTwoStepCloseMode() ? 0 : 1);
		startPerspectiveCombo.select(StartPerspectives.from(globalAppManager.getStartPerspective()).ordinal());
	}

	@Override
	public boolean performOk() {
		// Nur weiter machen, wenn denn auch wirklich dieser Dialog angezeigt worden war (= Objekte angelegt wurden)
		if (appLanguageCombo == null)
			return super.performOk();

		Locale	selectedLanguage	= eclipseINI.getLanguage();
		if (appLanguageComboSelection > -1 && appLanguageComboSelection != AvailableLanguages.fromLocale(eclipseINI.getLanguage()).ordinal()) {
			selectedLanguage	= AvailableLanguages.values()[appLanguageComboSelection].getLocale();
			eclipseINI.setLanguage(selectedLanguage);
		}

		if (1024 * maxHeapSpinner.getSelection() / 10 != eclipseINI.getMaxHeap()) {
			eclipseINI.setMaxHeap(1024 * maxHeapSpinner.getSelection() / 10);
			eclipseINI.setMinHeap(1024 * maxHeapSpinner.getSelection() / 10 / 8);
		}

		globalAppManager.setTwoStepCloseMode(twoStepCloseCombo.getSelectionIndex() == 0);
		globalAppManager.setStartPerspective(StartPerspectives.from(startPerspectiveCombo.getText()).getPerspectiveType());

		if (restartHintLabel.getVisible()) {
			try {
				eclipseINI.persist();

				String	restartQuestionTitle	= Messages.getString("STARTUP_SETTINGS_RESTART_QUESTION_TITLE", selectedLanguage);
				String	restartQuestionMessage	= Messages.getString("STARTUP_SETTINGS_RESTART_QUESTION_MESSAGE", selectedLanguage);
				if (MessageDialog.openQuestion(getShell(), restartQuestionTitle, restartQuestionMessage)) {
					workbench.restart();
				}
			} catch (IOException e) {
				showAccessDeniedWarning(e, selectedLanguage);
			}
		}

		return super.performOk();
	}

	private void showAccessDeniedWarning(IOException e, Locale language) {
		MessageDialog	dialog	= new MessageDialog(display.getActiveShell(),									// Shell
				Messages.getString("STARTUP_SETTINGS_WARNING_DIALOG_TITLE", language),							// Title
				null,																							// Image
				Messages.getString("STARTUP_SETTINGS_WARNING_DIALOG_MESSAGE", language).replace("%1", e.getCause().toString()),		// Message
				MessageDialog.WARNING,																			// Exclamation Mark
				new String[] {	Messages.getString("STARTUP_SETTINGS_WARNING_DIALOG_COPY_BUTTON", language),	// Button "Copy"
								IDialogConstants.OK_LABEL},														// Button "OK"
				1) {																							// Default-Button = OK

			@Override
			protected Control createCustomArea(Composite parent) {
				Text		iniContent	= new Text(parent, SWT.MULTI | SWT.BORDER);
				GridData	textLayout	= new GridData(SWT.FILL, SWT.FILL, true, true);
				iniContent.setLayoutData(textLayout);
				iniContent.setText(e.getMessage());
				return parent;
			}

			@Override
			protected void buttonPressed(int buttonId) {
				if (buttonId == 0) {
					Clipboard	clipboard	= new Clipboard(display);
					clipboard.setContents(new Object [] {e.getMessage()},
					                      new Transfer [] {TextTransfer.getInstance()});
					clipboard.dispose();
				} else {
					super.buttonPressed(buttonId);
				}
			}
		};
		ContextInjectionFactory.inject(dialog, context);
		dialog.open();
	}

}
